"""
artssat.io
========

The `artssat.io` module provides routines for the storing of simulations results.
"""
from netCDF4 import Dataset
from copy import copy
import numpy as np
import os

from artssat.sensor import ActiveSensor, PassiveSensor

class OutputFile:
    """
    Class to store results from an ARTS simulation to a NetCDF file.

    """
    def __init__(self,
                 filename,
                 dimensions = None,
                 mode = "a",
                 inputs = [],
                 floating_point_format = "f4",
                 full_retrieval_output = True):
        """
        Create output file to store simulation output to.

        Arguments:

            filename(str): Path of the output file.
            dimensions(list): List of tuples :code:`(s, e, o)` containing
                for each dimensions over which simulations will be produced
                the start index :code:`s`, the end index :code:`e` and
                the offset :code:`o`.
            mode(str): String describing the mode to open the output file.
            inputs: List of tuples :code:`(name, (dim1, ...))` containing the
                names (:code:`name`) of the input variables and names
                (:code:`dim1, ...`) of the associated dimensions.
            floating_point_format(str): The precision to use to store floating
                point numbers (:code:`f4` or :code:`f8`)
            full_retrieval_output(:code:`bool`): Whether or not to include
                full retrieval output (Jacobians, AVK and covariance matrices).
        """
        try:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            size = self.comm.Get_size()
            self.mpi = size > 1
        except:
            self.mpi = False

        filename = os.path.expanduser(filename)
        self.filename   = filename
        self.mode       = mode
        self.dimensions = dimensions
        self.f_fp       = floating_point_format
        self.inputs = inputs

        self.full_retrieval_output = full_retrieval_output

        if os.path.isfile(self.filename):
            if mode == "wb":
                os.remove(self.filename)
                self._initialized = False
            else:
                self._initialized = True
        else:
            self._initialized = False


    @property
    def initialized(self):
        return self._initialized

    def _initialize_dimensions(self):
        """
        Initialize compute dimension in the output file.
        """
        # indices
        for n, s, _ in self.dimensions:
            if s < 0:
                self.file_handle.createDimension(n, None)
            else:
                self.file_handle.createDimension(n, s)

    def _initialize_forward_simulation_output(self, simulation):
        """
        Initializes output file for forward simulation results.
        Create dict attribute :code:`variables` containing variables
        corresponding to each sensor name.
        """
        root = self.file_handle
        indices = [n for n, _, _ in self.dimensions]

        for s in simulation.sensors:
            dims = []
            if s.views > 1:
                dim = s.name + "_views"
                root.createDimension(dim, s.sensor_position.shape[0])
                dims += [dim]
                root.createVariable(s.name + "_position",
                                    self.f_fp,
                                    dimensions = tuple(indices + dims))
                root.createVariable(s.name + "_line_of_sight",
                                    self.f_fp,
                                    dimensions = tuple(indices + dims))

            if isinstance(s, ActiveSensor):
                dim = s.name + "_range_bins"
                root.createDimension(dim, s.range_bins.size - 1)
                dims += [dim]
            elif isinstance(s, PassiveSensor):
                dim = s.name + "_channels"
                root.createDimension(dim, s.f_grid.size)
                dims += [dim]
            if s.stokes_dimension > 1:
                dim = s.name + "_stokes_dim"
                root.createDimension(dim, s.stokes_dimension)
                dims += [dim]

            root.createVariable("y_" + s.name, self.f_fp,
                                dimensions = tuple(indices  + dims))

    def _initialize_retrieval_output(self, simulation):
        """
        Initialize output file for results from a retrieval
        calculation.

        Arguments:

            simulation: ArtsSimulation object from which to store the results.

        """

        retrieval = simulation.retrieval
        args      = simulation.args
        kwargs    = simulation.kwargs

        #
        # Global dimensions
        #

        # z grid
        p = simulation.workspace.p_grid.value
        self.file_handle.createDimension("z", p.size)

        # oem diagnostices
        self.file_handle.createDimension("oem_diagnostics", 5)

        #
        # Result groups
        #

        self.groups = []
        indices = [n for n, _, _ in self.dimensions]

        if not type(retrieval.results) == list:
            results = [retrieval.results]
        else:
            results = retrieval.results

        for r in results:
            group = self.file_handle.createGroup(r.name)
            self.groups += [group]

            # Retrieval quantities.
            for rq in retrieval.retrieval_quantities:
                v = group.createVariable(rq.name, self.f_fp, dimensions = tuple(indices + ["z"]))

            # OEM diagnostics.
            v = group.createVariable("diagnostics", self.f_fp,
                                     dimensions = tuple(indices + ["oem_diagnostics"]))

            # Observations and fit.
            for s in r.sensors:
                i, j = r.sensor_indices[s.name]
                m = j - i
                d1 = s.name + "_channels"
                group.createDimension(d1, m)
                v = group.createVariable("y_" + s.name, self.f_fp,
                                              dimensions = tuple(indices  + [d1]))
                v = group.createVariable("yf_" + s.name, self.f_fp,
                                              dimensions = tuple(indices  + [d1]))

            if self.full_retrieval_output:

                n = simulation.workspace.x.value.size
                m = simulation.workspace.y.value.size

                group.createDimension("m", m)
                group.createDimension("n", n)

                group.createVariable("G", self.f_fp, dimensions = tuple(indices + ["n", "m"]))
                group.createVariable("A", self.f_fp, dimensions = tuple(indices + ["n", "n"]))
                group.createVariable("covmat_so", self.f_fp, dimensions = tuple(indices + ["n", "n"]))
                group.createVariable("covmat_ss", self.f_fp, dimensions = tuple(indices + ["n", "n"]))
                group.createVariable("jacobian", self.f_fp, dimensions = tuple(indices + ["m", "n"]))

    def _initialize_inputs(self, simulation):

        self.input_dimensions = {}
        indices = [n for n, _, _ in self.dimensions]
        group = self.file_handle.createGroup("inputs")

        for (v, dims) in self.inputs:

            # Get data from provider.
            fget = getattr(simulation.data_provider, "get_" + v)
            data = fget(*simulation.args, **simulation.kwargs)
            if type(data) == str:
                data = np.array([d for d in data], dtype="S1")

            if not len(data.squeeze().shape) == len(dims):
                raise ValueError("Shape of input data {} does not match "
                                 " expected dimensions {}.".format(data.shape,
                                                                   dims))

            # Check if size of dimension has been inferred and if so if
            # it is consistent with previously inferred size.
            # Otherwise create dimension.

            for d, s in zip(dims, data.shape):
                if d in self.input_dimensions:
                    si = self.input_dimensions[d]
                    if si != s:
                        raise Exception("Dimension {} of input {} is inconsistent "
                                        "with inferred dimension ({})."
                                        .format(d, v, si))
                else:
                    group.createDimension(d, s)
                    self.input_dimensions[d] = s

            group.createVariable(v, data.dtype, tuple(indices + list(dims)))

        args   = [a - o for a, (_, _, o) in zip(simulation.args, self.dimensions)]
        kwargs = simulation.kwargs

    def initialize(self, simulation):
        """
        Initialize output file.

        This creates all necessary dimensions and variables in the NetCDF4
        output file. This function is run automatically before the first
        entry is stored in the file.
        """
        self.file_handle = Dataset(self.filename,
                                   mode = self.mode,
                                   parallel = self.mpi)

        try:
            self._initialize_dimensions()

            if len(simulation.retrieval.retrieval_quantities) > 0:
                self._initialize_retrieval_output(simulation)
            else:
                self._initialize_forward_simulation_output(simulation)
            if self.inputs:
                self._initialize_inputs(simulation)
        except:
            raise
        finally:
            self._initialized = True
            self.close()

    def _store_forward_simulation_results(self, simulation):

        args   = [a - o for a, (_, _, o) in zip(simulation.args, self.dimensions)]
        kwargs = simulation.kwargs

        for s in simulation.sensors:
            var = self.file_handle["y_" + s.name]
            y   = s.y
            var.__setitem__(list(args) + [slice(0, None)], y)
            if s.views > 1:
                var = self.file_handle[s.name + "_line_of_sight"]
                var.__setitem__(list(args) + [slice(0, None)],
                                s.sensor_line_of_sight.ravel())
                var = self.file_handle[s.name + "_position"]
                var.__setitem__(list(args) + [slice(0, None)],
                                s.sensor_position.ravel())


    def _store_retrieval_results(self, simulation):

        retrieval = simulation.retrieval
        args   = [a - o for a, (_, _, o) in zip(simulation.args, self.dimensions)]
        kwargs = simulation.kwargs

        if not type(retrieval.results) == list:
            results = [retrieval.results]
            groups = [self.file_handle.groups[retrieval.results.name]]
        else:
            results = retrieval.results
            groups = [self.file_handle.groups[r.name] for r in results]

        for g, r in zip(groups, results):

            #
            # Retrieved quantities
            #

            for rq in retrieval.retrieval_quantities:
                x = r.get_result(rq, interpolate = True, transform_back = True)
                if x is None:
                    x = r.get_xa(rq, interpolate = True, transform_back = True)
                var = g.variables[rq.name]
                var.__setitem__(list(args) + [slice(0, None)], x)

            #
            # OEM diagnostics.
            #
            var = g.variables["diagnostics"]
            var.__setitem__(list(args) + [slice(0, None)], r.oem_diagnostics)

            #
            # Observation and fit.
            #

            for s in r.sensors:
                i, j = r.sensor_indices[s.name]
                y  = r.y[i : j]
                yf = r.yf[i : j]

                name = "y_" + s.name
                var = g[name]
                var.__setitem__(list(args) + [slice(0, None)], y)

                name = "yf_" + s.name
                var = g[name]
                try:
                    var.__setitem__(list(args) + [slice(0, None)], yf)
                except:
                    print("Error storing yf: ", r.yf.shape, i, j)
                    pass

            #
            # Remaining retrieval output
            #

            if self.full_retrieval_output and r.oem_diagnostics[0] <= 2.0:
                ws = simulation.workspace

                var = g["A"]
                var.__setitem__(list(args) + [slice(0, None)] * 2, ws.avk.value)

                var = g["G"]
                var.__setitem__(list(args) + [slice(0, None)] * 2, ws.dxdy.value)

                var = g["covmat_ss"]
                var.__setitem__(list(args) + [slice(0, None)] * 2, ws.covmat_ss.value)

                var = g["covmat_so"]
                var.__setitem__(list(args) + [slice(0, None)] * 2, ws.covmat_so.value)

                var = g["jacobian"]
                var.__setitem__(list(args) + [slice(0, None)] * 2, ws.jacobian.value)

    def _store_inputs(self, simulation):

        if self.inputs:
            input_variables = self.file_handle["inputs"]
        else:
            return None

        for (v, dims) in self.inputs:
            # Get data from provider.
            fget = getattr(simulation.data_provider, "get_" + v)
            data = fget(*simulation.args, **simulation.kwargs)
            if type(data) == str:
                data = np.array([d for d in data], dtype="S1")

            args   = [a - o for a, (_, _, o) in zip(simulation.args, self.dimensions)]

            var = input_variables[v]
            if len(dims) > 0:
                var.__setitem__(list(args) + [slice(0, None)], data)
            else:
                var.__setitem__(list(args), data)

    def store_results(self, simulation):

        # Initialize file structure
        if not self.initialized:
            self.initialize(simulation)

        self.open()

        if len(simulation.retrieval.retrieval_quantities) > 0:
            self._store_retrieval_results(simulation)
        else:
            self._store_forward_simulation_results(simulation)
        if self.inputs:
            self._store_inputs(simulation)

        self.close()

    def open(self):
        if self.initialized and not self.mpi:
            if hasattr(self, "file_handle"):
                if not self.file_handle.isopen():
                    self.file_handle = Dataset(self.filename, mode = "r+")
            else:
                self.file_handle = Dataset(self.filename, mode = "r+")

    def close(self):
        if not self.mpi:
            if hasattr(self, "file_handle"):
                self.file_handle.close()

    def __getstate__(self):
        state = copy(self.__dict__)
        state.pop("file_handle", None)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)



r"""
The Abel-Boutle (2012) PSD
==========================

The Abel-Boutle (2012) PSD is a single moment PSD intended to represent rain
drops. Particle number densities are represented using a gamma distribution function

.. math::

    N(D) &= N_0\ D^\gamma \ \exp(-\lambda D).

The parameters :math:`N_0` and :math:`\lambda` can be diagnosed from the rain water
content using

.. math::

    \lambda &= \left [ \frac{\pi \rho_\text{w} x_1 \Gamma(4 + \mu)}{6 \rho_\text{air} q_\text{R}}]^{\frac{1}{4 + \mu - x_2}}
    N_0 &= x_1 \lambda^{x_2}


.. [AB2012] Abel SJ, Boutle IA. 2012. An improved representation of the raindrop size distribution for
single-moment microphysics schemes. Q. J. R. Meteorol. Soc. 138: 2151–2162. DOI:10.1002/qj.1949

"""
import numpy as np
import scipy as sp
from scipy.special import gamma
from pyarts.workspace import arts_agenda

from artssat import dimensions as dim
from artssat.scattering.psd.data.psd_data import D_eq
from artssat.scattering.psd.arts.arts_psd import ArtsPSD
from artssat.scattering.psd.data.psd_data import PSDData



class AB12(ArtsPSD):
    r"""
    The AB12 class provides an implementation of the Abel-Boutle (2012) single-moment
    PSD for rain drops.
    """

    @classmethod
    def from_psd_data(self, psd, mu = 0.0):
        r"""
        Create a AB12 PSD from given psd data.

        Parameters:

            psd(PSDData or other PSD): PSD data from which to create the MY05
            representation.
            mu(:code:`float` or array): The value of the mu parameter to use.
        """
        mass_density   = psd.get_mass_density()
        return AB12(mu, mass_density)

    def __init__(self,
                 mu = 0.0,
                 mass_density = None):
        r"""
        Parameters:
            mu(:code:`numpy.float`): The :math:`\mu` parameter of the PSD
            mass_density(:code:`numpy.ndarray`): Array containing
            the water content for a given set of volume elements in an
            atmosphere.
        """
        self.mu = mu

        if not mass_density is None:
            self.mass_density = mass_density

        super().__init__(D_eq(1000.0))

    def convert_from(self, psd):
        r"""
        Convert given psd to AB12 PSD with :math:`\mu` parameter of this instance.

        Parameters:

            psd: Other PSD providing :code:`get_moment` and :code:`get_mass_density`
            member functions.
        """
        self.mass_density   = psd.get_mass_density()

    def _get_parameters(self):
        """
        Checks if parameters of the PSD are available and tries to broadcast
        them to the shape of the mass density data. Calculates parameters of

        Returns:

            :code:`tuple(n0, lmbd, mu)` containing the  parameters of
            the PSD function.

        Raises:

            An exception if any of the parameters is not set or cannot be
            broadcasted into the shape of the number density data.
        """

        # Number density

        # Mass density
        m = self.mass_density
        if m is None:
           raise Exception("The mass density needs to be set to use"
                            " this function.")
        shape = m.shape

        try:
            mu = np.broadcast_to(np.array(self.mu), shape)
        except:
            raise Exception("Could not broadcast mu paramter to the shape"
                            "of the provided intercept parameter N.")


        x1 = 0.22
        x2 = 2.2
        lmbd = (np.pi * 1000.0 * x1 * gamma(4 + mu)) / (6.0 *  m)
        lmbd = lmbd ** (1.0 / (4.0 + mu - x2))
        print(lmbd)
        n0 = x1 * lmbd ** x2

        return n0, lmbd, mu

    @property
    def moment_names(self):
        """
        The names of the predictive moments of the PSD.
        """
        return ["mass_density"]

    @property
    def moments(self):
        try:
            return [self.mass_density]
        except:
            return None

    @property
    def pnd_call_agenda(self):
        """
        The ARTS SM implementing the MY05 PSD.
        """
        @arts_agenda
        def pnd_call(ws):
            ws.psdAbelBoutle12(t_min = self.t_min,
                               t_max = self.t_max)
        return pnd_call

    def get_moment(self, p, reference_size_parameter = None):
        r"""
        Analytically computes the :math:`p` th moment :math:`M(p)` of the PSD
        using

        .. math::

            M(p) = N_0  \lambda^{-(p + \mu + 1)}
                   \Gamma (p + 1 + \mu)

        Parameters:

            p(:code:`float`): Which moment of the distribution to compute.

        Returns:

            :code:`numpy.ndarray` containing the :math:`p` th moment for
            all volume elements described by the PSD.

            reference_size_parameter(:class: `SizeParameter`): Size parameter
            with respect to which the moment should be computed.

        """
        if not reference_size_parameter is None:
            a1 = self.size_parameter.a
            b1 = self.size_parameter.b
            a2 = reference_size_parameter.a
            b2 = reference_size_parameter.b

            c = (a1 / a2) ** (p / b2)
            p = p * b1 / b2
        else:
            c = 1.0

        n0, lmbd, mu = self._get_parameters()
        m = n0 * lmbd ** (-(mu + p + 1)) * gamma(mu + 1.0 + p)

        m[lmbd == 0.0] = 0.0

        return c * m

    def get_mass_density(self):
        r"""
        Returns:

             The :code:`numpy.ndarray` containing the mass density data of
             the PSD.

        """
        return self.mass_density

    def evaluate(self, x):
        r"""
        Compute a numeric representation of the PSD data.

        Parameters:

            x(:code:`numpy.ndarray`): Array containing the values of the size
            parameter at which to evaluate the PSD.

        Returns:

            :code:`PSDData` object containing the numeric PSD data corresponding
            to this PSD.

        """
        n0, lmbd, mu = self._get_parameters()

        shape = n0.shape
        result_shape = shape + (1,)

        n0   = np.reshape(n0, result_shape)
        lmbd = np.broadcast_to(lmbd, shape).reshape(result_shape)
        mu   = np.broadcast_to(mu, shape).reshape(result_shape)

        x = x.reshape((1,) * len(shape) + (-1,))

        y = n0 * x ** mu * np.exp(- lmbd * x)
        return PSDData(x, y, self.size_parameter)



"""
The :code:`scattering` subpackage provides classes of scattering species
and particle size distributions (PSDs) for the simulation of scattering
in an ARTS simulation.

In addition to that, the :code: `artssat.scattering.psd` subpackage
contains some functionality for the handling of PSD data.

"""
from artssat.scattering.scattering_species \
    import ScatteringSpecies
from artssat.scattering.psd \
    import D14, MY05
from artssat.scattering.solvers import RT4




"""
artssat.sensor.utils
--------------------

Helper functions for handling satellite sensors.
"""
import numpy as np
import scipy as sp

def sensor_properties(center_frequencies, offsets, order = "positive"):
    """
    Compute ARTS frequency grid array and sensor response matrix for
    for paired side-band channels from a list of given center frequencies
    and corresponding offsets.

    Args:

        center_frequencies: List of center frequencies of the sensor

        offsets: List of lists of offsets for each sensor frequency

        order: "negative" if the sum of center frequencies and negative
            offset should be used to order the frequencies. Positive
            otherwise.
    """
    i = 0 # output index, increased only for <order> offsets.
    j = 0 # f_grid index, increased for each offset

    f_grid = []
    sr_data = []
    sr_i    = []
    sr_j    = []
    ci = 0

    for f, ofs in zip(center_frequencies, offsets):
        ofs.sort()

        nfs = len(ofs)

        for i, o in enumerate(ofs[::-1]):
            if o > 0.0:
                f_grid  += [f - o]
                sr_data += [0.5]
                sr_j    += [j]
                if order == "negative":
                    sr_i    += [ci + i]
                else:
                    sr_i += [ci + nfs - 1 - i]
                j += 1

        if ofs[0] == 0.0:
            f_grid  += [f]
            sr_data += [1.0]
            sr_j    += [j]
            if order == "negative":
                sr_i += [ci + i]
            else:
                sr_i += [ci]
            j += 1

        for i, o in enumerate(ofs):
            if o > 0.0:
                f_grid  += [f + o]
                sr_data += [0.5]
                sr_j    += [j]
                if order == "negative":
                    sr_i    += [ci + nfs - 1 - i]
                else:
                    sr_i += [ci + i]
                j += 1

        ci += nfs
    sensor_response = sp.sparse.coo_matrix((sr_data, (sr_i, sr_j)))

    return np.array(f_grid), sensor_response



"""
Binned
======

The binned PSD represents a discretized PSD that is represented
by particle densities over a given number of bins.
"""
import numpy as np
import scipy as sp
from pyarts.workspace import arts_agenda

from artssat.scattering.psd.arts.arts_psd import ArtsPSD
from artssat.scattering.psd.data.psd_data import PSDData, D_eq

class Binned(ArtsPSD):
    """
    Discrete representation of a PSD using fixed bins.
    """

    def convert_from(self, psd):
        """
        Convert another psd to a Binned object.
        """
        self.size_parameter = psd.size_parameter
        y = psd.evaluate(self.x).data
        inds = (slice(None),) * (len(y.shape) - 1)
        self.moments = [y[inds + (i,)] for i in range(self.x.size)]

    def __init__(self,
                 x,
                 size_parameter = D_eq(1000.0),
                 t_min = 0.0,
                 t_max = 400.0):
        """
        Arguments:

            x: Vector holding the positions of the centers of the size bins.

            size_parameter: The size parameter to use.
        """
        super().__init__(size_parameter)
        self.x = x.ravel()
        self.t_min = t_min
        self.t_max = t_max


    def get_moment(self, p, reference_size_parameter = None):
        if not reference_size_parameter is None:
            a1 = self.size_parameter.a
            b1 = self.size_parameter.b
            a2 = reference_size_parameter.a
            b2 = reference_size_parameter.b

            c = (a1 / a2) ** (p / b2)
            p = p * b1 / b2
        else:
            c = 1.0

        data = np.array([m for m in self.moments]).T
        x = np.broadcast_to(self.x.reshape(1, -1), data.shape)
        return c * np.trapz(data * x ** p, x = x)

    @property
    def moment_names(self):
        """
        Names of the moments of the PSD.

        These are just :code:`bin_0, bin_1, ...` for each bin
        of the PSD.
        """
        return ["bin_" + str(i) for i in range(self.x.size)]

    @property
    def pnd_call_agenda(self):
        @arts_agenda
        def pnd_call(ws):
            ws.Ignore(ws.pnd_agenda_input)
            ws.Ignore(ws.pnd_agenda_input_t)
            ws.Ignore(ws.pnd_agenda_input_names)
            ws.Ignore(ws.dpnd_data_dx_names)
            ws.Touch(ws.psd_data)
            ws.Touch(ws.dpsd_data_dx)

            ws.scat_species_a = self.size_parameter.a
            ws.scat_species_b = self.size_parameter.b

            xi = ws.psd_size_grid.value
            y  = ws.pnd_agenda_input.value
            y  = np.maximum(y, 1e-12)
            yi = sp.interpolate.interp1d(self.x.ravel(),
                                         np.log10(y), axis = 1, fill_value = 0.0)(xi)
            yi = 10.0 ** yi
            yi[yi == 1e-12] == 0.0

            t = ws.pnd_agenda_input_t.value
            yi[t < self.t_min, :] = 0.0
            yi[t > self.t_max, :] = 0.0
            yi = np.maximum(yi, 0.0)

            ws.psd_data.value = np.copy(yi)
            ndx = len(ws.dpnd_data_dx_names.value)
            ws.dpsd_data_dx = np.ones((ndx, yi.shape[0], self.x.size))

        return pnd_call



import numpy as np

import threading

from bokeh                               import events
from bokeh.colors                        import RGB
from bokeh.server.server                 import Server
from bokeh.application                   import Application
from bokeh.application.handlers.function import FunctionHandler
from bokeh.plotting                      import figure, ColumnDataSource, curdoc
from bokeh.models.widgets                import Div, Tabs, Panel, Select, CheckboxGroup, \
    CheckboxButtonGroup, RadioButtonGroup
from bokeh.models.glyphs                 import Image
from bokeh.models                        import Range1d, LinearColorMapper, ColorBar, \
    BasicTicker, PrintfTickFormatter, FuncTickFormatter, Rect
from bokeh.models.callbacks              import CustomJS
from bokeh.layouts                       import column, row
from bokeh.palettes                      import RdBu, Inferno

from artssat.sensor import ActiveSensor, PassiveSensor


class BokehServer():
    """
    Handles the singleton Bokeh server used to serve visualisations
    if no jupyter notebook is open.
    """
    def __init__(self):
        self.server = None

    def __del__(self):
        if not self.server is None:
            self.server.io_loop.stop()
            self.server.stop()

    def start(self, application, port = 5000):

        if not self.server is None:
            self.server.io_loop.stop()
            self.server.stop()

        self.server = Server(application, port = port)
        self.server.io_loop.start()


standalone_server = BokehServer()

################################################################################
# Profile plot
################################################################################

class ProfilePlot:
    """
    A profile plot displays the height distribution of an atmospheric quantity
    with respect to either altitude or pressure.
    """

    def __init__(self, z, p, title, name, width = 500):

        def update_y_variable(attrname, old, new):

            if self.y_axis_quantity.active == 0:
                self.y = self.z
                self.y_range = Range1d(start = self.z[0], end = self.z[-1])

            if self.y_axis_quantity.active == 1:
                self.y = self.p
                self.y_range = Range1d(start = self.p[-1], end = self.p[0])

            for s in self.sources:
                s.data["y"] = self.y

            fig = curdoc().get_model_by_name(self.name).children[0]
            fig.y_range = self.y_range

        def update_x_scale(attrname, old, new):
            if self.x_axis_scale.active == 1:
                for s in self.sources:
                    s.data["x"] = np.log10(s.data["x_"])
            else:
                for s in self.sources:
                    s.data["x"] = s.data["x_"]

        def update_y_scale(attrname, old, new):
            if self.y_axis_scale.active == 1:
                for s in self.sources:
                    s.data["y"] = np.log10(self.y)
            else:
                for s in self.sources:
                    s.data["y"] = self.y

        self.z = np.copy(z)
        self.p = np.copy(p)
        self.y = z
        self.quantities = []

        self.name = name

        self.x_axis_scale = RadioButtonGroup(labels = ["linear", "log"],
                                             active = 0)
        self.x_axis_scale.on_change("active", update_x_scale)

        self.y_axis_scale = RadioButtonGroup(labels = ["linear", "log"],
                                             active = 0)
        self.y_axis_scale.on_change("active", update_y_scale)

        self.y_axis_quantity = RadioButtonGroup(labels = ["altitude", "pressure"],
                                                active = 0)
        self.y_axis_quantity.on_change("active", update_y_variable)

        self.sources = []

        self.fig = figure(title = title, width = width, height = 500)

    def add_quantity(self, x, line_kwargs = {}):
        self.sources += [ColumnDataSource(data = dict(x = x,
                                                      x_ = x,
                                                      y = self.y))]
        self.fig.line("x", "y", source = self.sources[-1], **line_kwargs)

    @property
    def doc(self):
        d1 = Div(text = "y-axis variable:", width = 200)
        d2 = Div(text = "x-axis scale:", width = 200)
        d3 = Div(text = "y-axis scale:", width = 200)
        return column(self.fig,
                      row(d1, self.y_axis_quantity),
                      row(d2, self.x_axis_scale),
                      row(d3, self.y_axis_scale),
                      name = self.name)

class RetrievalResultPlot:
    def __init__(self, z, p, retrieval):
        self.z = z
        self.p = p

        y = z

        self.retrieval = retrieval
        self.retrieval_quantities = retrieval.retrieval_quantities

        self.figures = dict([(rq, ProfilePlot(self.z,
                                              self.p,
                                              rq.name,
                                              "rq_plot_" + rq.name,
                                              width = 400)) for
                             rq in self.retrieval_quantities])
        self.status = dict([(rq, True) for rq in self.retrieval_quantities])
        self.sources_x  = dict([(rq, []) for rq in self.retrieval_quantities])
        self.sources_xa = dict([(rq, []) for rq in self.retrieval_quantities])
        self.sources_x0 = dict([(rq, []) for rq in self.retrieval_quantities])
        self.lines_x  = dict([(rq, []) for rq in self.retrieval_quantities])
        self.lines_xa = dict([(rq, []) for rq in self.retrieval_quantities])
        self.lines_x0 = dict([(rq, []) for rq in self.retrieval_quantities])

        #
        # Get retrieval results as list.
        #
        if type(self.retrieval.results) is list:
            results = self.retrieval.results
        else:
            results = [self.retrieval.results]

        self.colors = Inferno[len(results) + 2][1:-1]

        #
        # Add plots for each retrieval quantity.
        #

        for i, rr in enumerate(self.retrieval.results):
            for rq in self.retrieval_quantities:
                xa = rr.get_result(rq, attribute = "xa")
                x  = rr.get_result(rq, attribute = "x")
                x0 = rr.get_result(rq, attribute = "x0")

                if xa is None:
                    continue

                self.sources_x[rq]  += [ColumnDataSource(data =
                                                         dict(x = x, y = y))]
                self.sources_xa[rq] += [ColumnDataSource(data =
                                                         dict(x = xa, y = y))]
                self.sources_x0[rq] += [ColumnDataSource(data =
                                                         dict(x = x0, y = y))]

                fig = self.figures[rq]
                fig.add_quantity(x, line_kwargs = dict(line_color = self.colors[i],
                                                       line_width = 2))
                if i == 0:
                    fig.add_quantity(xa, line_kwargs = dict(line_dash = "dashed",
                                                            line_color = self.colors[i],
                                                            line_width = 2))
                #fig.add_quantity(x0, line_kwargs = dict(line_dash  = "dashdot",
                #                                        line_color = self.colors[i],
                #                                        line_width = 2))


        self.plots = row(*[self.figures[k].doc for k in self.figures],
                         name = "retrieval_quantities")

        def update_plots(attrname, old, new):
            state = self.checks.active
            plots = curdoc().get_model_by_name('retrieval_quantities')

            print(state)
            for i, rq in enumerate(self.retrieval_quantities):
                if i in state:
                    if self.status[rq] == True:
                        continue
                    else:
                        self.status[rq] = True
                        plots.children.append(self.figures[rq].doc)
                else:
                    if self.status[rq] == False:
                        continue
                    else:
                        fig = curdoc().get_model_by_name("rq_plot_" + rq.name)
                        self.status[rq] = False
                        plots.children.remove(fig)

            print(plots.children)
            print(self.status)


        #
        # Checkbox button to hide plots
        #

        labels  = [rq.name for rq in self.retrieval.retrieval_quantities]
        active   = list(range(len(labels)))
        self.checks = CheckboxButtonGroup(labels = labels, active = active)
        self.checks.on_change("active", update_plots)


    def make_doc(self):
        title = Div(text = "<h2>Retrieved quantities</h2>")
        c = column(title, self.checks, self.plots)
        return c

class AVKPlot:
    def __init__(self, simulation):
        self.simulation = simulation
        self.sources = {}

        if not type(simulation.retrieval.results) is list:
            result = simulation.retrieval.results
        else:
            result = simulation.retrieval.results[-1]

        self.figures = {}
        self.images  = {}

        rqs = simulation.retrieval.retrieval_quantities
        self.panels = [self._make_avk_plot(result, rq) for rq in rqs]



    def _make_avk_plot(self, result, rq):
        avk = result.get_avk(rq)
        x   = np.arange(avk.shape[0])
        xx, yy  = np.meshgrid(x, x)
        self.sources[rq] = ColumnDataSource(data = dict(x = xx.ravel(),
                                                        y = yy.ravel(),
                                                        z = avk.ravel()))
        self.figures[rq] = figure(x_range = (x[0] - 0.5, x[-1] + 0.5),
                                  y_range = (x[0] - 0.5, x[-1] + 0.5),
                                  tools = ["box_zoom", "box_select", "reset"],
                                  width = 600,
                                  aspect_scale = 1.0,
                                  toolbar_location = "above")

        colors = RdBu[11]
        clim_high = max(avk.max(), -avk.min())
        clim_low  = min(-avk.max(), avk.min())

        mapper = LinearColorMapper(palette= colors,
                                   low = clim_low,
                                   high= clim_high)
        self.images[rq] = self.figures[rq].rect(source = self.sources[rq],
                                                x = "x",
                                                y = "y",
                                                fill_color = {'field' : 'z',
                                                              'transform' : mapper},
                                                line_color = None,
                                                width = 1,
                                                height = 1)
        #marker_source = ColumnDataSource(data = dict(x = np.array([x[0]]),
        #                                             y = np.array([x[0]])))

        #marker = Rect(x = "x",
        #              y = "y",
        #              fill_color = None,
        #              line_color = RGB(0.0, 0.0, 0.0),
        #              width = x[-1] - x[0],
        #              height = 1)
        #self.images[rq].add_glyph(marker_source, rect)

        #axis = self.figures[rq].xaxis
        #axis.major_label_overrides = dict(zip(axis.ticker,
        #                                      z[axis.ticker]))
        #axis = self.figures[rq].yaxis
        #axis.major_label_overrides = dict(zip(axis.ticker,
        #                                      z[axis.ticker]))

        def update(attr, old, new):
            print(old)
            print(new)
            print(self.sources[rq].selected.indices)

        #
        # The colorbar
        #

        color_bar = ColorBar(color_mapper = mapper,
                             #major_label_text_font_size ="5pt",
                             ticker = BasicTicker(desired_num_ticks = len(colors)),
                             formatter = PrintfTickFormatter(format="%.2f"),
                             label_standoff = 6,
                             border_line_color=None,
                             location=(0, 0))
        self.figures[rq].add_layout(color_bar, 'right')

        #
        # Single altitude
        #


        source = ColumnDataSource(data = dict(x = x, y = avk[0]))
        fig =  figure(tools = ["box_zoom", "box_select", "reset"],
                      width = 400,
                      y_range = [clim_low, clim_high],
                      toolbar_location = "above")
        line = fig.line(x = "x", y = "y", source = source)


        #
        # Plot callback
        #

        cb = CustomJS(args = dict(source_avk = self.sources[rq],
                                  #source_marker = marker_source,
                                  source = source), code =
                      """
                      var x = Number(cb_obj["x"]);
                      var y = Number(cb_obj["y"]);
                      console.log(x, y);
                      var i = Math.floor(y);
                      var j = i + 1;
                      var n = source.data["x"].length;
                      source.data["y"] = [];
                      for (var k = 0; k < n; k++) {
                          source.data["y"].push(source_avk.data["z"][i * n + k]);
                      }
                      source.change.emit();

                      console.log(source.data["x"].length);
                      console.log(source.data["y"].length);
                      """)
        self.figures[rq].js_on_event(events.Tap, cb)
        r = row(self.figures[rq], fig)

        return Panel(child = r, title = rq.name, width = 500)

    def make_doc(self):
        title = Div(text = "<h2>Averaging kernels</h2>")
        t = Tabs(tabs = self.panels)
        return column(title, t)

class ObservationPlot:

    def __init__(self, sensor):

        self.sensor = sensor
        self.observations = []

    def make_active_sensor_plot(self):

        fig = figure(title = self.sensor.name, width = 500)

        z = self.sensor.range_bins
        x = self.sensor.y
        z = np.ravel(0.5 * (z[1:] + z[:-1]))

        if self.observations == []:
            observations = [sensor.y]
        else:
            observations = self.observations

        for o in observations:
            x = o.reshape(z.size, -1)
            y = z // 1e3
            fig.line(x = x, y = o)
            fig.circle(x = x, y = o)

        fig.xaxis.axis_label = "Radar reflectivity [{0}]"\
                            .format(self.sensor.iy_unit)
        fig.yaxis.axis_label = "Altitude [km]"

        return fig

    def make_passive_sensor_plot(self):

        fig = figure(title = self.sensor.name, width = 500)

        x = self.sensor.f_grid
        y = self.sensor.y

        for o in self.observations:
            fig.line(x = x, y = o)
            fig.circle(x = x, y = o)

        return fig


    def add_observation(self, o):
        self.observations += [o]

    def make_doc(self):

        if isinstance(self.sensor, ActiveSensor):
            return self.make_active_sensor_plot()
        else:
            return self.make_passive_sensor_plot()


################################################################################
# Atmosphere
################################################################################


def plot_temperature(simulation, data_provider):

    fig = figure(title = "Temperature", width = 500)

    t = simulation.workspace.t_field.value.ravel()
    z = simulation.workspace.z_field.value.ravel() / 1e3

    fig.line(x = t, y = z)
    fig.circle(x = t, y = z)

    fig.xaxis.axis_label = "Temperature [K]"
    fig.yaxis.axis_label = "Altitude [km]"

    return fig

def plot_absorption_species(simulation, data_provider):

    fig = figure(title = "Absorption species", width = 500)

    for a in simulation.atmosphere.absorbers:

        i = a._wsv_index
        x = simulation.workspace.vmr_field.value[i, :, :, :].ravel()
        z = simulation.workspace.z_field.value.ravel() / 1e3

        fig.line(x = x, y = z)
        fig.circle(x = x, y = z)

    fig.xaxis.axis_label = "VMR [mol/m^3]"
    fig.yaxis.axis_label = "Altitude [km]"

    return fig


def plot_scattering_species(simulation, data_provider):
    pass


def make_atmosphere_panel(simulation, data_provider):
    c = column(plot_temperature(simulation, data_provider),
               plot_absorption_species(simulation, data_provider))
    return Panel(child = c, title = "Atmosphere", width = 600)

################################################################################
# Retrieval
################################################################################

def plot_temperature(simulation, data_provider):

    fig = figure(title = "Temperature", width = 500)

    t = simulation.workspace.t_field.value.ravel()
    z = simulation.workspace.z_field.value.ravel() / 1e3

    fig.line(x = t, y = z)
    fig.circle(x = t, y = z)

    fig.xaxis.axis_label = "Temperature [K]"
    fig.yaxis.axis_label = "Altitude [km]"

    return fig

def plot_absorption_species(simulation, data_provider):

    fig = figure(title = "Absorption species", width = 500)

    for a in simulation.atmosphere.absorbers:

        i = a._wsv_index
        x = simulation.workspace.vmr_field.value[i, :, :, :].ravel()
        z = simulation.workspace.z_field.value.ravel() / 1e3

        fig.line(x = x, y = z)
        fig.circle(x = x, y = z)

    fig.xaxis.axis_label = "VMR [mol/m^3]"
    fig.yaxis.axis_label = "Altitude [km]"

    return fig


def plot_scattering_species(simulation, data_provider):
    pass


def make_atmosphere_panel(simulation, data_provider):
    c = column(plot_temperature(simulation, data_provider),
               plot_absorption_species(simulation, data_provider))
    return Panel(child = c, title = "Atmosphere", width = 600)

################################################################################
# Sensor
################################################################################

def make_sensor_plot(sensor):

    fig = figure(title = sensor.name, width = 500)

    if isinstance(sensor, ActiveSensor):
        z = sensor.range_bins
        x = sensor.y
        z = 0.5 * (z[1:] + z[:-1])

        for i in range(sensor.y.shape[1]):
            x = sensor.y[:, i]
            y = z.ravel() // 1e3
            fig.line(x = x, y = y)
            fig.circle(x = x, y = y)

        fig.xaxis.axis_label = "Radar reflectivity [{0}]"\
                            .format(sensor.iy_unit)
        fig.yaxis.axis_label = "Altitude [km]"

    if isinstance(sensor, PassiveSensor):

        x = sensor.f_grid
        y = sensor.y

        for i in range(sensor.y.shape[1]):
            fig.line(x = x, y = y[:, i])
            fig.circle(x = x, y = y[:, i])


    return fig


def make_sensor_panel(sensors):
    c = column(*[make_sensor_plot(s) for s in sensors],
               sizing_mode = "fixed")
    return Panel(child = c, title = "Measurements", width = 600)

def make_retrieval_panel(simulation):

    z = np.copy(simulation.workspace.z_field.value.ravel())
    p = np.copy(simulation.workspace.p_grid.value.ravel())

    retrieval_result_plot = RetrievalResultPlot(z, p, simulation.retrieval)
    r = retrieval_result_plot.make_doc()

    if type(simulation.retrieval.results) is list:
        results = simulation.retrieval.results
    else:
        results = [simulation.retrieval.results]

    if all([not r.avk is None for r in simulation.retrieval.results]):
        avks = AVKPlot(simulation)
        r.children += avks.make_doc().children

    return Panel(child = r, title = "Retrieval", width = 600)


    observation_plots = {}
    for s in simulation.sensors:
        observation_plots[s] = ObservationPlot(s)

    if type(simulation.retrieval.results) is list:

        plots = []
        for r in simulation.retrieval.results:

            for s in r.sensors:
                i, j = r.sensor_indices[s.name]
                observation_plots[s].add_observation(r.y[i : j])
                print(r.y[i:j])
                observation_plots[s].add_observation(r.yf[i : j])
                print(r.yf[i:j])

            plot = ProfilePlot(z, p)
            for q in r.retrieval_quantities:
                x = r.get_result(q)
                plot.add_quantity(x, r.name, dict(line_dash = "solid",
                                                  line_width = 3))
                xa = r.get_result(q, "xa")
                plot.add_quantity(xa, r.name, dict(line_dash = "dashed"))
                x0 = r.get_result(q, "x0")
                if not all(xa == x0):
                    plot.add_quantity(x0, r.name, dict(line_dash = "dotted"))

            plots += [plot]
    else:
        plot = ProfilePlot(z, p)
        for q in r.retrieval_quantities:
            x = r.get_result(q)
            plot.add(x, r.name, "blue")
        plots = [plot]

    title = Div(text = "<h2>Retrieved quantities</h2>")
    doms = [title] + [p.make_doc() for p in plots]

    title = Div(text = "<h2>Fitted measurements</h2>")
    doms += [title] + [observation_plots[s].make_doc() for s in simulation.sensors]

    c     = column(*doms)
    return Panel(child = c, title = "Retrieval", width = 600)

def make_document_factory(simulation):

    def make_document(doc):

        doc.title = "parts dashboard"

        sensor_p     = make_sensor_panel(simulation.sensors)
        atmosphere_p = make_atmosphere_panel(simulation, simulation.data_provider)
        retrieval_p  = make_retrieval_panel(simulation)



        t = Tabs(tabs = [sensor_p,
                         atmosphere_p,
                         retrieval_p])
        doc.add_root(t)

    return make_document

def dashboard(simulation):
    apps = {'/': Application(FunctionHandler(make_document_factory(simulation)))}
    standalone_server.start(apps, port = 8880)




r"""
The modified gamma distribution PSD
===================================

The form of the modified gamma distribution (MGD) used here is as follows:

.. math::

    \frac{N(X)}{dX} = N \frac{\nu}{\Gamma(1 + \alpha)}\lambda^{\nu(1 + \alpha)}
                        D^{\nu(1 + \alpha) - 1} \cdot \exp \{-(\lambda D)^\nu\}.

The distribution is described by four parameters:

    1. The intercept parameter :math:`N`
    2. The slope parameter :math:`\lambda`
    3. The shape parameter :math:`\alpha`
    4. The parameter :math:`\nu`

"""
import numpy as np
import scipy as sp
from scipy.special import gamma
from artssat import dimensions as dim
from artssat.scattering.psd.arts.arts_psd import ArtsPSD
from artssat.scattering.psd.data.psd_data import PSDData, D_eq

class ModifiedGamma(ArtsPSD):
    r"""
    The :class:`ModifiedGamma` class describes the size distribution of
    scattering particles in an atmosphere using the four parameters of the
    particle size distribution.
    """

    properties = [("intercept_parameter", (dim.p, dim.lat, dim.lon), np.ndarray),
                  ("alpha", (dim.p, dim.lat, dim.lon), np.ndarray),
                  ("lmbd", (dim.p, dim.lat, dim.lon), np.ndarray),
                  ("nu", (dim.p, dim.lat, dim.lon), np.ndarray)]

    def __init__(self,
                 size_parameter,
                 intercept_parameter = None,
                 alpha = None,
                 lmbd = None,
                 nu = None):
        r"""
        Create instance of the modified gamma distribution with given parameters.

        If any of the parameters is neither provided and nor explicitly set
        afterwards, it will be requested from the data provider. However, most
        operations on PSDs will require the values to be set and can thus first
        be performed when the object has access to the data.

        Parameters:
            size_parameter(SizeParameter): The SizeParameter instance describing
            the size parameter that should be used fo PSD.

            intercept_parameter(numpy.float or ndarray): The intercept parameter
            :math:`N`

            alpha(numpy.float or ndarray): The shape parameter :math:`\alpha`.
                Must be broadcastable into the shape of N.

            lmbd(numpy.float or ndarray): The slope parameter :math:`\lambda`.
                Must be broadcastable into the shape of N.

            nu(numpy.float or ndarray): The :math:`\nu` parameter. Must be
                broadcastable into the shape of N.
        """
        if not intercept_parameter is None:
            self.intercept_parameter = intercept_parameter

        shape = self.intercept_parameter.shape

        if not alpha is None:
            try:
                self.alpha = np.broadcast_to(alpha, shape)
            except:
                raise Exception("Could not broadcast alpha parameter to shape "
                                "of intercept parameter.")

        if not lmbd is None:
            try:
                self.lmbd = np.broadcast_to(lmbd, shape)
            except:
                raise Exception("Could not broadcast lambda parameter to shape "
                                " of intercept parameter.")

        if not nu is None:
            try:
                self.nu = np.broadcast_to(nu, shape)
            except:
                raise Exception("Could not broadcast nu parameter to shape "
                                " of N parameter.")

        super().__init__(size_parameter)

    def _get_parameters(self):
        """
        Checks if parameters of the PSD are available and tries to broadcast
        them to the shape of the intercept parameter.

        Returns:

            :code:`tuple(n, alpha, lmbd, nu)` containing the four parameters of
            the PSD.

        Raises:

            An exception if any of the MGD parameters is not set or cannot be
            broadcasted.
        """

        n = self.intercept_parameter
        if n is None:
           raise Exception("The intercept parameter needs to be set to use"
                            " this function.")
        shape = n.shape

        # Lambda parameter

        lmbd = self.lmbd
        if lmbd is None:
            raise Exception("The lambda parameter needs to be set to use "
                            "this function")
        try:
            lmbd = np.broadcast_to(lmbd, shape)
        except:
            raise Exception("Could not broadcast lambda paramter to the shape"
                            "of the provided intercept parameter N.")

        # Alpha parameter

        alpha = self.alpha
        if alpha is None:
            raise Exception("The alpha parameter needs to be set to use "
                            "this function.")
        try:
            alpha = np.broadcast_to(alpha, shape)
        except:
            raise Exception("Could not broadcast alpha paramter to the shape"
                            "of the provided intercept parameter N.")

        # Nu parameter

        nu = self.nu
        if nu is None:
            raise Exception("The nu parameter needs to be set to use this"
                            "function.")

        try:
            nu = np.broadcast_to(nu, shape)
        except:
            raise Exception("Could not broadcast nu paramter to the shape"
                            "of the provided intercept parameter N.")

        return n, lmbd, alpha, nu

    @property
    def moment_names(self):
        r"""
        The free parameters of the PSD.
        """
        return []

    def get_moment(self, p, reference_size_parameter = None):
        r"""
        Computes the :math:`p` th moment :math:`M(p)` of the PSD using

        .. math::
            M(p) = \frac{N}{\lambda} \frac{\Gamma (1 + \alpha + p / \nu )}
            {\Gamma({1 + \alpha})}.

        Parameters:
            p(np.float): Which moment of the PSD to compute.

        Raises:
            Exception: If any of the parameters of the PSD is not set.

        """
        if not reference_size_parameter is None:
            a1 = self.size_parameter.a
            b1 = self.size_parameter.b
            a2 = reference_size_parameter.a
            b2 = reference_size_parameter.b

            c = (a1 / a2) ** (p / b2)
            p = p * b1 / b2
        else:
            c = 1.0

        n, lmbd, alpha, nu = self._get_parameters()

        m = n / lmbd ** p
        m *= gamma(1 + alpha + p / nu)
        m /= gamma(1 + alpha)

        return c * m

    def get_mass_density(self):
        r"""
        Computes the mass density :math: `\rho_m` for the given bulk elements
        using

        .. math::
            \rho_m = a \cdot M(b).

        where :math:`a` and :math:`b` are the coefficients of the mass-size
        relation of the size parameter.

        Returns:

            :code:`numpy.ndarray` containing the mass density corresponding
            to each volume element described by the PSD.
        """
        a = self.size_parameter.a
        b = self.size_parameter.b
        return a * self.get_moment(b)

    @property
    def pnd_call_agenda(self):
        r"""
        ARTS agenda that contains the call to the WSM that computed this PSD.
        """

        n0 = np.nan
        if not self.intercept_parameter is None \
           and self.intercept_parameter.size == 1:
            n0 = self.intercept_parameter[0]

        mu = np.nan
        if not self.mu is None \
           and self.mu.size == 1:
            mu = self.mu[0]

        lmbd = np.nan
        if not self.lmbd is None \
           and self.lmbd.size == 1:
            lmbd = self.lmbd[0]

        nu = np.nan
        if not self.nu is None \
           and self.nu.size == 1:
            nu = self.nu[0]

        @arts_agenda
        def pnd_call(ws):
            ws.psdMgd(n0 = n0,
                      mu = mu,
                      la = lambd,
                      gam   = nu,
                      t_min  = self.t_min,
                      t_max  = self.t_max)

        return pnd_call

    def evaluate(self, x):
        r"""
        Computes the values of this modified gamma distribution evaluated at
        the given size grid :code:`x`.

        Parameters:

            x(numpy.array): Array containing the values of the size parameter
            at which to evaluate the PSD.

        Returns:

            :class:`PSDData` object containing the numeric PSD data obtained
            by evaluating this PSD at the given values of the size parameter.

        """
        n, lmbd, alpha, nu = self._get_parameters()

        shape = n.shape

        n     = n.reshape(shape + (1,))
        lmbd  = lmbd.reshape(shape  + (1,))
        alpha = alpha.reshape(shape  + (1,))
        nu    = nu.reshape(shape + (1,))

        print(n, lmbd, alpha, nu)

        y = n * nu / gamma(1 + alpha)
        y *= lmbd ** (nu * (1.0 + alpha))
        y = y * x ** (nu * (1.0 + alpha) - 1) \
             * np.exp(- (lmbd * x) ** nu)
        return PSDData(x, y, self.size_parameter)



r"""
The Milbrandt-Yau (2005) PSD
============================

The Milbrandt-Yau (2005) PSD used in the GEM model uses a modified
gamma distribution of the form:

.. math::
    N(D_{max}) = N_0 D_{max}^\nu \exp(- \lambda D_{max} ^ \mu)

In this two-moment scheme, the number density :math:`\rho_n` and mass
density :math:`\rho_m`  are used as predictive moments. They can be related
to the :math:`N_0` and :math:`\lambda` parameters of the PSD using

.. math::
    \rho_n = \frac{N_0}{\mu} \lambda^{\frac{\nu}{\mu}}
             \Gamma (\frac{\nu + 1}{\mu})

    \rho_m = a\frac{N_0}{\mu} \lambda^{\frac{\nu + b}{\mu}}
             \Gamma ( \frac{\nu + 1 + b}{\mu} ),

where :math:`a, b` are the parameters of the mass-size relationship

.. math::
    M(D) = a \cdot D_{max}^b.

"""
import numpy as np
import scipy as sp
from scipy.special import gamma
from pyarts.workspace import arts_agenda

from artssat import dimensions as dim
from artssat.scattering.psd.modified_gamma     import ModifiedGamma
from artssat.scattering.psd.data.psd_data      import D_max
from artssat.scattering.psd.arts.arts_psd import ArtsPSD
from artssat.scattering.psd.data.psd_data import PSDData



settings = {"cloud_ice" : (0.0, 1.0, 440., 3.0),
            "rain" : (0.0, 1.0, 523.5988, 3.0),
            "snow" : (0.0, 1.0, 52.35988, 3.0),
            "graupel" : (0.0, 1.0, 209.4395, 3.0),
            "hail" : (0.0, 1.0, 471.2389, 3.0),
            "cloud_water" : (1.0, 1.0, 523.5988, 3.0)}

class MY05(ArtsPSD):
    r"""
    The :class:`MY05` class describes the size distributions of particles
    in an atmosphere using the number density :math:`\rho_n` and mass
    density :math:`\rho_m` as predictive moments.

    The :math:`\nu` and :math:`\mu` parameters take on fixed values depending
    on the hydrometeor type.
    """

    properties = [("number_density", (dim.p, dim.lat, dim.lon), np.ndarray),
                  ("mass_density", (dim.p, dim.lat, dim.lon), np.ndarray),
                  ("nu", (), np.float),
                  ("mu", (), np.float)]

    @classmethod
    def from_psd_data(self, psd, nu, mu, a, b):
        r"""
        Create a MY05 PSD from given psd data.

        Parameters:

            psd(PSDData or other PSD): PSD data from which to create the MY05
            representation.

            nu(:code:`float`): :math:`\nu` parameter of MY05 PSD

            mu(:code:`float`): :math:`\mu` parameter of MY05 PSD

            a(:code:`float`): :math:`a` coefficient of the mass-size relationship

            b(:code:`float`): :math:`b` coefficient of the mass-size relationship

            hydrometeor_type: One of ["cloud_ice", "rain", "snow", "graupel", "hail"]
                or None. If not None this will override the given settings for
                nu, mu, a, b.
        """
        size_parameter = D_max(a, b)
        number_density = psd.get_moment(0)
        mass_density   = psd.get_mass_density()

        return MY05(nu, mu, a, b, None, number_density, mass_density)

    def __init__(self,
                 nu = None,
                 mu = None,
                 a  = None,
                 b  = None,
                 hydrometeor_type = None,
                 number_density = None,
                 mass_density = None):
        r"""
        Parameters:
            nu(:code:`numpy.float`): The :math:`\nu` parameter of the PSD

            mu(:code:`numpy.float`): The :math:`\mu` parameter of the PSD

            a(:code:`numpy.float`): The :math:`a` coefficient of the
            mass-size realtionship.

            b(:code:`numpy.float`): The :math:`b` coefficient of the
            mass-size realtionship.

            number_density(:code:`numpy.ndarray`): Array containing
            the number density for a given set of volume elements in an
            atmosphere.

            mass_density(:code:`numpy.ndarray`): Array containing
            the mass density for a given set of volume elements in an
            atmosphere.
        """
        if not hydrometeor_type is None:
            if hydrometeor_type in settings:
                nu, mu, a, b = settings[hydrometeor_type]
            else:
                raise Exception("Expected keyword hydrometeor type to be one of"
                                " {0} but got '{1}'.".format(list(settings.keys()),
                                                           hydrometeor_type))

        self.nu = nu
        self.mu = mu

        self.hydrometeor_type = hydrometeor_type

        if not number_density is None:
            self.number_density = number_density

        if not mass_density is None:
            self.mass_density = mass_density

        super().__init__(D_max(a, b))

    def convert_from(self, psd):
        r"""
        Convert given psd to MY05 PSD with :math:`\nu,\mu,a` and :math:`b`
        parameters of this instance.

        Parameters:

            psd: Other PSD providing :code:`get_moment` and :code:`get_mass_density`
            member functions.
        """
        self.number_density = psd.get_moment(0)
        self.mass_density   = psd.get_mass_density()

    def _get_parameters(self):
        """
        Checks if parameters of the PSD are available and tries to broadcast
        them to the shape of the number density data. Converts the mass density
        and number density data to the :math:`N_0` and :math:`\lambda`
        parameters of the PSD function.

        Returns:

            :code:`tuple(n0, lmbd, mu, nu)` containing the four parameters of
            the PSD function.

        Raises:

            An exception if any of the parameters is not set or cannot be
            broadcasted into the shape of the number density data.
        """

        # Number density
        n = self.number_density
        if n is None:
           raise Exception("The number density needs to be set to use"
                            " this function.")
        shape = n.shape

        # Mass density
        m = self.mass_density
        if m is None:
           raise Exception("The mass density needs to be set to use"
                            " this function.")
        try:
            m = np.broadcast_to(m, shape)
        except:
            raise Exception("Could not broadcast mass density to the shape"
                            "of the provided intercept parameter N.")

        # Alpha parameter
        try:
            mu = np.broadcast_to(self.mu, shape)
        except:
            raise Exception("Could not broadcast alpha paramter to the shape"
                            "of the provided intercept parameter N.")

        # Nu parameter
        try:
            nu = np.broadcast_to(self.nu, shape)
        except:
            raise Exception("Could not broadcast nu paramter to the shape"
                            "of the provided intercept parameter N.")

        a = self.size_parameter.a
        b = self.size_parameter.b

        lmbd = (a * n) / m \
               * gamma((nu + 1 + b) / mu) \
               / gamma((nu + 1) / mu)
        lmbd = lmbd ** (mu / b)

        inds = np.logical_or(n == 0.0, m == 0.0)
        lmbd[inds] = 0.0

        n0 = n * mu * lmbd ** ((nu + 1.0) / mu) / gamma((nu + 1.0) / mu)
        n0[inds] = 0.0

        return n0, lmbd, mu, nu

    @property
    def moment_names(self):
        """
        The names of the predictive moments of the PSD.
        """
        return ["number_density", "mass_density"]

    @property
    def moments(self):
        try:
            return [self.number_density, self.mass_density]
        except:
            return None

    @property
    def pnd_call_agenda(self):
        """
        The ARTS WSM implementing the MY05 PSD.
        """
        @arts_agenda
        def pnd_call(ws):
            ws.psdMilbrandtYau05(hydrometeor_type = self.hydrometeor_type,
                                 t_min = self.t_min,
                                 t_max = self.t_max)
        return pnd_call

    def get_moment(self, p, reference_size_parameter = None):
        r"""
        Analytically computes the :math:`p` th moment :math:`M(p)` of the PSD
        using

        .. math::

            M(p) = \frac{N_0}{\mu} \lambda ^{-\frac{\nu + p + 1}{\mu}}
                   \Gamma ( \frac{\nu + p + 1}{\mu})

        Parameters:

            p(:code:`float`): Which moment of the distribution to compute.

        Returns:

            :code:`numpy.ndarray` containing the :math:`p` th moment for
            all volume elements described by the PSD.

            reference_size_parameter(:class: `SizeParameter`): Size parameter
            with respect to which the moment should be computed.

        """
        if not reference_size_parameter is None:
            a1 = self.size_parameter.a
            b1 = self.size_parameter.b
            a2 = reference_size_parameter.a
            b2 = reference_size_parameter.b

            c = (a1 / a2) ** (p / b2)
            p = p * b1 / b2
        else:
            c = 1.0

        n0, lmbd, mu, nu = self._get_parameters()
        m = n0 / mu * lmbd ** (-(nu + p + 1) / mu) * gamma((nu + 1.0 + p) / mu)

        m[lmbd == 0.0] = 0.0

        return c * m

    def get_mass_density(self):
        r"""
        Returns:

             The :code:`numpy.ndarray` containing the mass density data of
             the PSD.

        """
        a = self.size_parameter.a
        b = self.size_parameter.b
        return a * self.get_moment(b)

    def evaluate(self, x):
        r"""
        Compute a numeric representation of the PSD data.

        Parameters:

            x(:code:`numpy.ndarray`): Array containing the values of the size
            parameter at which to evaluate the PSD.

        Returns:

            :code:`PSDData` object containing the numeric PSD data corresponding
            to this PSD.

        """
        n0, lmbd, mu, nu = self._get_parameters()

        shape = n0.shape
        result_shape = shape + (1,)

        n0   = np.reshape(n0, result_shape)
        lmbd = np.broadcast_to(lmbd, shape).reshape(result_shape)
        mu   = np.broadcast_to(mu, shape).reshape(result_shape)
        nu   = np.broadcast_to(nu, shape).reshape(result_shape)

        x = x.reshape((1,) * len(shape) + (-1,))

        y = n0 * x ** nu * np.exp(- lmbd * x ** mu)
        return PSDData(x, y, self.size_parameter)



"""
ARTS PSD interface
==================

The ARTS interface is implemented by the :class:`ArtsPSD` class, which
implements common functionality for the handling of PSDs in ARTS and defines
the interaction with the :class:`ArtsSimulation` and in particular
the :class:`ScatteringSpecies` instance in which the PSD is used.
"""




"""
artssat.retrieval
-----------------

The :code:`retrieval` module handles retrieval calculations with ARTS.
The retrieval functionality is provided by a class structure that is
similar to the one used for Jacobians:

1. :class:`RetrievalCalculation` handles the actual calculation of
   retrievals through the ARTS workspace

2. :class:`RetrievalQuantity` defines the general interface for quantities
   which can be retrieved and how to add them to the list of quantities
   to retrieve.

3. :class:`RetrievalBase` defines an abstract base class for the
   retrieval classes that encapsulate the settings and code that
   is specific to the given quantity.

Retrieving quantities
=====================

So that a quantity can be retrieved, it must inherit from the
:class:`RetrievalQuantity` base class. In this case, it can simply be added to
the :code:`retrieval` attribute of a given simulation:

::

    simulation.retrieval.add(q)

This will add the quantity :code:`q` to the quantities which will be retrieved.
Moreover it will instantiate the :class:`retrieval_class` class of the quantity
:code:`q` and set the :code:`q.retrieval` attribute of :code:`q`.

It is important to note that :code:`simulation.retrieval` and :code:`q.retrieval`
are of different types: :code:`simulation.retrieval` is of type
:code:`RetrievalCalculation` and coordinates the interaction with the ARTS workspace
required for the retrieval. The attribute :code:`q.retrieval` of the retrieval quantity
holds the settings and retrieval results specific to the quantity :code:`q`.

Reference
=========
"""

import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import weakref

from abc import ABCMeta, abstractmethod, abstractproperty

from pyarts.workspace         import arts_agenda
from pyarts.workspace.agendas import Agenda
from pyarts.workspace.methods import workspace_methods
wsm = workspace_methods

from artssat.jacobian    import JacobianBase, JacobianQuantity, Transformation
from artssat.arts_object import ArtsObject, arts_property
from artssat.arts_object import Dimension as dim
from artssat.sensor      import ActiveSensor, PassiveSensor


################################################################################
# RetrievalBase
################################################################################

class RetrievalBase(ArtsObject, metaclass = ABCMeta):
    """
    Base class for quantity-specific retrieval classes.

    The quantity specific retrieval classes are meant to hold settings as well
    as retrieval results after a calculation.

    In order to retrieve any quantity the following must be given:

        1. The a priori mean state :code:`xa` of the retrieval quantity
           of the Gaussian distribution representing the a priori assumptions.

        2. A covariance matrix :code:`covariance` or its inverse,
           a precision matrix :code:`precision`, of the Gaussian
           distribution representing the a priori assumption on the retrieval
           quantities. It is also possible to specify both, in which case
           the user has to make sure that they are consistent.

    In addition the following attributes can be given to customize the
    behaviour of the retrieval:

        3. An initial state :code:`x0` to start the retrieval iteration.

        4. A lower limit :code:`limit_low` which is used in each iteration
           as a lower cutoff for the next iteration step $x_i$.

        5. An upper limit :code:`limit_high` which is used in each iteration
           as a lower cutoff for the next iteration step $x_i$.

    .. note: All numeric values given for a retrieval quantity are assumed
        to be in the transformed space of the quantity. If a quantity has
        constant a priori mean :math:`x_a = 10^{-5}` and is
        :math:`log_{10}`-transformed, :code:`xa` should be given as :code:`-5`.

    """
    @arts_property(["Sparse, Matrix"], shape = (dim.Joker, dim.Joker), optional = True)
    def covariance_matrix(self):
        """
        The covariance matrix for the retrieval quantity.

        Each quantity must have at least one of the :code:`covariance_matrix`
        or the :code:`precision_matrix` attributes set.
        """
        return None

    @arts_property(["Sparse, Matrix"], shape = (dim.Joker, dim.Joker), optional = True)
    def precision_matrix(self):
        """
        The inverse of the covariance matrix.

        Each quantity must have at least one of the :code:`covariance_matrix`
        or the :code:`precision_matrix` attributes set.
        """
        return None

    @arts_property("Vector", shape = (dim.Joker,))
    def xa(self):
        """
        The mean of the Gaussian a priori distribution assumed for the
        quantity.
        """
        return None

    @arts_property("Vector", shape = (dim.Joker,), optional = True)
    def x0(self):
        """
        Optional start value for the retrieval iteration.
        """
        return None

    @arts_property("Numeric", optional = True)
    def limit_low(self):
        """
        Optional lower cutoff to apply to an iteration state :math:`x_i` before
        performing the  the forward simulation and also the backward
        transformation of the retrieval quantity.
        """
        return None

    @arts_property("Numeric", optional = True)
    def limit_high(self):
        """
        Optional upper cutoff to apply to an iteration state :math:`x_i` before
        performing the  the forward simulation and also the backward
        transformation of the retrieval quantity.
        """
        return None

    def get_xa(self, data_provider, *args, **kwargs):
        """
        Get a priori vector from data provider and set the :code:`xa`
        attribute this object.

        Arguments:

            data_provider: Data provide from which to query a priori
                state.

            *args, **kwargs: Arguments and keyword arguments to be passed
                on to the data provider.
        """
        try:
            fname = "get_" + self.quantity.name + "_xa"
            xa_fun = getattr(data_provider, fname)
            xa = xa_fun(*args, **kwargs)
            self.xa = xa_fun(*args, **kwargs)

        except AttributeError:
            raise Exception("The data provider must provide a get method for "
                           "the a priori state of retrieval quantity {0}."
                           .format(self.quantity.name))

    def setup(self, ws, data_provider, *args, **kwargs):
        """
        Generic setup method for retrieval quantities.

        This method tries to get the a priori vector :math:`x_a`, the start vector
        :math:`x_0`, the covariance and the precision matrix from the data
        provider and sets them in the given workspace.

        If no start vector :math:`x_0` is provided, the a priori state is used
        as initial state for the retrieval iteration.

        Of the covariance or precision matrices only one must be specified.

        Arguments:

            ws(:class:`pyarts.workspace.Workspace`): The ARTS workspace on
                which to setup the retrieval.

            data_provider: Data provider object to query a priori settings from.

            :code:`*args` and :code:`**kwargs` are forwarded to the data provider.
        """

        #
        # Get a priori and start values.
        #

        self.get_xa(data_provider, *args, **kwargs)

        fname = "get_" + self.quantity.name + "_x0"
        if hasattr(data_provider, fname):
            x0_fun = getattr(data_provider, fname)
            self.x0 = x0_fun(*args, **kwargs)
        else:
            self.x0 = np.copy(self.xa)

        self.get_data(ws, data_provider, *args, **kwargs)

        self.add(ws)

        #
        # Get covariance and precision matrix.
        #

        fname = "get_" + self.quantity.name + "_covariance"
        try:
            covmat_fun = getattr(data_provider, fname)
            covmat = covmat_fun(*args, **kwargs)
        except AttributeError as e:
            covmat = None
        except Exception as e:
            raise e

        fname = "get_" + self.quantity.name + "_precision"
        try:
            precmat_fun = getattr(data_provider, fname)
            precmat = precmat_fun(*args, **kwargs)
        except AttributeError as e:
            precmat = None
        except Exception as e:
            raise e

        if covmat is None and precmat is None:
            raise Exception("The data provider must provide a get method for "
                            "the covariance or the precision matrix of retrieval"
                            "quantity {0}." .format(self.quantity.name))

        self.quantity.transformation.setup(ws, data_provider, *args, **kwargs)

        if not covmat is None:
            ws.covmat_sxAddBlock(block = covmat)
        if not precmat is None:
            ws.covmat_sxAddInverseBlock(block = precmat)

    def get_iteration_preparations(self, index):

        if self.limit_low is None and self.limit_high is None:
            return None

        limit_low = -np.inf
        if self._limit_low.fixed:
            limit_low = self._limit_low.value

        limit_high = np.inf
        if self._limit_high.fixed:
            limit_high = self._limit_high.value

        def agenda(ws):
            ws.xClip(ijq = index, limit_low = limit_low, limit_high = limit_high)

        return arts_agenda(agenda)

################################################################################
# RetrievalQuantity
################################################################################

class RetrievalQuantity(JacobianQuantity):
    """
    Abstract interface for quantities for which can be retrieved using ARTS.

    Instances of the :code:`RetrievalQuantity` class can be added to the
    retrieval quantities of a simulation, which means that artssat will try
    to retrieve their value from a given observation.

    The interface for retrieval quantities is implemented by four properties
    that are shared by all quantities that can be retrieved with ARTS:

    1. :code:`transformation`: Retrieval quantities can be retrieved not only
        in the units used inside ARTS but also as transformed quantities, e.g.
        in :math:`log_{10}` space. The transformation property hols for each
        retrieval quantity a :class:`artssat.jacobian.Transformation` object
        representing the applied transformation.

    2. :code:`fixed`: Property indicating whether a retrieval quantity should
        be held fixed for a given retrieval run.

    3. :code:`retrieval_class`: A python class object holding the
        quantity-specific retrieval class that is instantiated and set as the
        quantities' :code:`retrieval` attribute when the quantity is added as
        retrieval quantity.

    4. :code:`retrieval`: Instance of the quantities :code:`retrieval_class`
        holding the retrieval settings and results specific to the quantity
        and a given retrieval run.
    """

    def __init__(self):
        self._fixed          = None
        self._retrieval      = None
        super().__init__()

    @property
    def fixed(self):
        """
        Whether or not the retrieval quantity should be held fixed for a given
        retrieval run.
        """
        return self._fixed

    @fixed.setter
    def fixed(self, t):
        self._fixed = t

    @abstractmethod
    def set_from_x(self, ws, xa):
        """
        Set method that sets the value of the retrieval quantity to a given
        a priori state.
        """
        pass

    @abstractproperty
    def retrieval_class(self):
        """
        Return the class object that holds the actual implementation of the
        Jacobian calculation.
        """
        pass

    @property
    def retrieval(self):
        """
        The :code:`jacobian_class` object holding the quantity-specific settings
        and actual results of the Jacobian calculations for this quantity.
        """
        return self._retrieval

    @retrieval.setter
    def retrieval(self, r):
        if not isinstance(r, self.retrieval_class):# or \
           #not isinstance(r, RetrievalBase):
            raise ValueError("The retrieval property of a RetrievalQuantity"\
                             " can only be set to an instance of the objects"\
                             "own retrieval_class.")
        else:
            self._retrieval = r


################################################################################
# RetrievalCalculation
################################################################################

class RetrievalRun:
    """
    The :class:`RetrievalRun` represents a single call to the ARTS OEM
    workspace method. A single retrieval calculation can consist of
    several sequential retrieval runs performed with different settings.
    Each of these runs is represented by a single :code:`RetrievalRun`
    instance that holds the retrieval settings as well as the results
    of the single run.

    Attributes:

        name(:code:`str`): A name to identity the retrieval run.

        simulation(:code:`artssat.ArtsSimulation`): The simulation instance
            on which this retrieval run is executed.

        y(:code:`y`): The observation vector for the retrieval.

        settings(:code:`dict`): A dictionary holding the retrieval
            settings.
    """
    def __init__(self,
                 name,
                 simulation,
                 y,
                 settings,
                 sensor_indices,
                 retrieval_quantities,
                 previous_run = None):

        for rq in retrieval_quantities:
            rq.fixed = False

        self.name           = name
        self.y              = np.copy(y)
        self.settings       = settings
        self.sensors        = simulation.sensors.copy()
        self.sensor_indices = sensor_indices
        self.rq_indices     = {}
        self.retrieval_quantities = retrieval_quantities.copy()
        self.previous_run   = previous_run

        self._simulation = weakref.ref(simulation)
        self._data_provider = weakref.ref(simulation.data_provider)

        self.x = None

    @property
    def simulation(self):
        simulation = self._simulation()
        if simulation:
            return simulation
        else:
            raise ValueError("The corresponding simulation has "
                                    " been deleted.")

    @property
    def data_provider(self):
        data_provider = self._data_provider()
        if data_provider:
            return data_provider
        else:
            raise ValueError("The corresponding data provider has "
                                    " been destroyed.")

    def get_result(self, q, attribute = "x", interpolate = False, transform_back = False):

        if q in self.retrieval_quantities:
            i, j = self.rq_indices[q]
            x = getattr(self, attribute)
            if x is None:
                return x
            x = x[i : j]

            if transform_back:
                x = q.transformation.invert(x)

            if interpolate:
                ws = self.simulation.workspace
                grids = [ws.p_grid.value, ws.lat_grid.value, ws.lon_grid.value]
                grids = [g for g in grids if g.size > 0]
                x = q.retrieval.interpolate_to_grids(x, grids)

            return x

        if not self.previous_run is None:
            return self.previous_run.get_result(q, attribute = attribute)
        else:
            return None

    def get_xa(self, q, interpolate = True, transform_back = False):

        if transform_back:
            xa = q.transformation.invert(q.retrieval.xa)
        else:
            xa = q.retrieval.xa

        if interpolate:
            ws = self.simulation.workspace
            grids = [ws.p_grid.value, ws.lat_grid.value, ws.lon_grid.value]
            grids = [g for g in grids if g.size > 0]
            xa = q.retrieval.interpolate_to_grids(xa, grids)

        return xa

    def get_avk(self, q):
        if q in self.retrieval_quantities:
            i, j = self.rq_indices[q]
            x = getattr(self, "avk")
            return x[i : j, i : j]

        if not self.previous_run is None:
            return self.previous_run.get_result(q, attribute = "avk")
        else:
            return None

    #
    # Plotting functions
    #

    def plot_result(self,
                    q,
                    ax = None,
                    transform_back = True,
                    include_prior = True,
                    data_provider = None,
                    args = [],
                    kwargs = {}):
        """
        Plot retrieved results of given quantity.

        Works only in 1-dimensional atmospheres.

        Args:
            q: The retrieval quantity of which to plot the results
            ax: matplotilib Axes object in which to plot the results. If
                 None, a new axes object is constructed using suplots(1, 1).
            transform_back: Whether or not to plot results in
                transformed (:code:`False`) or original state space
                (True, default)
            include_prior: If :code:`True` also the a priori mean is plotted
                using a dashed line.
        """

        x = self.get_result(q,
                            interpolate = True,
                            transform_back = transform_back)
        if x is None:
            s = "No result for retrieval quantity {} available.".format(q.name)
            raise Exception(s)

        if ax is None:
            _, ax = plt.subplots(1, 1)

        z = self.simulation.workspace.z_field.value.ravel()

        ls = ax.plot(x, z, label = q.name)

        if include_prior:
            xa = self.get_xa(q,
                             interpolate = True,
                             transform_back = transform_back)
            ax.plot(xa, z, label = q.name, c = ls[0].get_color(), ls = "--")

        if data_provider:
            getter = getattr(data_provider, "get_" + q.name)
            x = getter(*args, **kwargs)
            ax.plot(x, z, label = "Reference", c = ls[0].get_color(), ls = "-.")

        return ax

    def plot_jacobian(self,
                      sensor,
                      q,
                      ax = None):
        """
        Plot the Jacobian of observations from a given sensor w.r.t.
        a given retrieval quantity. Note that the Jacobian is always
        displayed in transformed coordinates.

        Works only in 1-dimensional atmospheres.

        Args:
            sensor: The sensor providing the osbervations of which the Jacobian
                should be plotted.
            q: The retrieval quantity w.r.t. to which the Jacobian should be plotted.
            ax: matplotilib Axes object in which to plot the results. If
                 None, a new axes object is constructed using suplots(1, 1).
        """
        i1, j1 = self.sensor_indices[sensor.name]
        i2, j2 = self.rq_indices[q]
        dydx = self.jacobian[i1 : j1, i2: j2]
        print(dydx.shape)

        if dydx is None:
            s = "No result for retrieval quantity {} available.".format(q.name)
            raise Exception(s)

        if ax is None:
            _, ax = plt.subplots(1, 1)

        for i in range(dydx.shape[0]):
            ax.plot(dydx[i, :], label = "Channel {}".format(i))
        return ax

    def plot_a_priori_errors(self, *args, **kwargs):
        """
        Plots contributions of different components of the x-vector to the OEM
        cost.
        """
        i = 0
        for q in self.retrieval_quantities:
            _, ax = plt.subplots(1, 1)

            xa = q.retrieval.xa
            x = self.get_result(q,
                                interpolate = False,
                                transform_back = False)
            dx = x - xa

            data_provider = self.simulation.data_provider
            fname = "get_" + q.name + "_covariance"
            try:
                covmat_fun = getattr(data_provider, fname)
                covmat = covmat_fun(*args, **kwargs)
            except AttributeError as e:
                covmat = None
            except Exception as e:
                raise e

            fname = "get_" + q.name + "_precision"
            try:
                precmat_fun = getattr(data_provider, fname)
                precmat = precmat_fun(*args, **kwargs)
            except AttributeError as e:
                precmat = None
            except Exception as e:
                raise e

            if precmat:
                c = dx * np.dot(precmat, dx)
            else:
                c = dx * np.linalg.solve(covmat, dx)

            ax.plot(c)
            ax.set_title(q.name, loc = "left")

    def plot_observation_errors(self):
        """
        Plots contributions of different components of the x-vector to the OEM
        cost.
        """
        data_provider = self.simulation.data_provider
        get_name = "get_observation_error_covariance"
        f = getattr(data_provider, get_name)
        covmat_se = f(*args, **kwargs)

        for s in self.sensors:
            _, ax = plt.subplots(1, 1)
            i, j = self.sensor_indices[s.name]
            y = self.y[i : j]
            yf = self.yf[i : j]
            dy = yf - y
            s = covmat_se[i : j, i : j]
            c = dy * np.dot(s, dy)
            ax.plot(c)
            ax.set_title(s.name, loc = "left")

    def setup_a_priori(self, *args, **kwargs):
        """
        Setup the retrieval calculation.

        This methods performs the setup necessary to run the retrieval
        calculations on the given workspace. That means it does:

        1. Gather a priori mean states and covariance matrices for all
            registered retrieval quantities.

        2. Define the retrieval inside the given ARTS workspace.

        3. Construct the ARTS inversion iterate agenda.

        Arguments:

            ws(:code:`arts`): The :code:`pyarts.workspace.Workspace`
                object which to setup for the retrieval.

            sensors(list): The sensors to use for the retrieval calculation.

            scattering(:code:`bool`): Whether or not to the forward calculations
                involve scattering.

            retrieval_provider: The data provider providing the a priori data
                for the retrieval.

            *args, **kwargs: Arguments and keyword arguments that are passed on
                to the data provider.
        """

        ws            = self.simulation.workspace
        data_provider = self.data_provider

        xa = []
        x0 = []

        ws.retrievalDefInit()

        #
        # Need to store indices of retrieval quantities in x
        # for future reference.
        #

        rq_index = 0

        for rq in self.retrieval_quantities:

                rq.retrieval.setup(ws, data_provider, *args, **kwargs)

                xa += [rq.retrieval.xa]

                if not self.previous_run is None:
                    r = self.previous_run.get_result(rq)
                else:
                    r = None

                if not r is None:
                    x0 += [r]
                else:
                    if rq.retrieval.x0 is None:
                        x0 += [rq.retrieval.xa]
                    else:
                        x0 += [rq.retrieval.x0]

                self.rq_indices[rq] = (rq_index, rq_index + xa[-1].size)
                rq_index += xa[-1].size

        #
        # Set values of retrieval quantities excluded from retrieval.
        #

        for rq in self.simulation.retrieval.retrieval_quantities:

            # Need to setup transformation in case RQ has not yet been
            # retrieved.
            t = rq.transformation
            if hasattr(t, "initialize"):
                t.initialize(data_provider, *args, **kwargs)

            # Get data for retrieval quantity.
            rq.retrieval.get_data(ws, data_provider, *args, **kwargs)

            # Get result from previous run (x_p).
            # If not available set to a priori.
            x_p = self.get_result(rq)
            if x_p is None:
                rq.retrieval.get_xa(data_provider, *args, **kwargs)
                x_p = rq.retrieval.xa
            rq.set_from_x(ws, x_p)

        ws.retrievalDefClose()

        self.xa = np.concatenate(xa)
        self.x0 = np.concatenate(x0)

        ws.xa = self.xa
        ws.x  = self.x0

        #
        # We extract from the observation error covariance matrix
        # the artssat that are required for the used sensors.
        #

        get_name = "get_observation_error_covariance"
        f = getattr(data_provider, get_name)
        covmat_se = f(*args, **kwargs)

        if sp.sparse.issparse(covmat_se):
            covmat_se = covmat_se.todense()

        covmat_blocks = []

        for s in self.sensors:
            if isinstance(s, ActiveSensor):
                (i, j) = self.sensor_indices[s.name]
                covmat_blocks += [covmat_se[i:j, i:j]]

        for s in self.sensors:
            if not isinstance(s, ActiveSensor):
                (i, j) = self.sensor_indices[s.name]
                covmat_blocks += [covmat_se[i:j, i:j]]

        covmat_se = sp.sparse.block_diag(covmat_blocks, format = "coo")
        ws.covmat_seAddBlock(block = covmat_se)

    def setup_iteration_agenda(self):

        debug = self.simulation.retrieval.debug_mode

        if debug:
            self.debug = {"x"               : [],
                          "yf"              : [],
                          "jacobian"        : [],
                          "iteration_index" : []}

        ws            = self.simulation.workspace
        data_provider = self.data_provider

        s = self.sensors[0]
        ws.Copy(ws.sensor_los,  s._wsvs["sensor_los"])
        ws.Copy(ws.sensor_pos,  s._wsvs["sensor_pos"])
        ws.Copy(ws.sensor_time, s._wsvs["sensor_time"])

        #
        # Setup inversion iterate agenda
        #

        agenda = Agenda.create("inversion_iterate_agenda")

        @arts_agenda
        def debug_print(ws):
            ws.Print(ws.x, 0)
        #agenda.append(debug_print)

        for i, rq in enumerate(self.retrieval_quantities):
            preps = rq.retrieval.get_iteration_preparations(i)
            if not preps is None:
                agenda.append(preps)

        arg_list = self.sensors[0].get_wsm_kwargs(wsm["x2artsAtmAndSurf"])
        agenda.add_method(ws, wsm["x2artsAtmAndSurf"], **arg_list)

        scattering = len(self.simulation.atmosphere.scatterers) > 0
        if scattering:
            agenda.add_method(ws, wsm["pnd_fieldCalcFromParticleBulkProps"])

        i_active = []
        i_passive = []
        for i,s in enumerate(self.sensors):
            if isinstance(s, ActiveSensor):
                i_active += [i]
            if isinstance(s, PassiveSensor):
                i_passive += [i]

        i = 0
        y_index = 0

        # Active sensor
        if len(i_active) > 0:
            s = self.sensors[i_active[0]]

            agenda.append(arts_agenda(s.make_y_calc_function(append = False)))

            i += 1

        atmosphere = self.simulation.atmosphere
        # Passive sensor
        for s in [self.sensors[i] for i in i_passive]:

            # Scattering solver call
            if scattering:
                scattering_solver = self.simulation.scattering_solver
                agenda.append(arts_agenda(scattering_solver.make_solver_call(atmosphere, s)))

            agenda.append(arts_agenda(
                s.make_y_calc_function(append = i > 0,
                                       scattering = scattering)
            ))
            i += 1


        def iteration_finalize(ws):
            ws.Ignore(ws.inversion_iteration_counter)

            ws.Copy(ws.yf, ws.y)
            ws.jacobianAdjustAndTransform()

        agenda.append(arts_agenda(iteration_finalize))

        @arts_agenda
        def get_debug(ws):
            self.debug["x"]               += [np.copy(ws.x.value)]
            self.debug["yf"]              += [np.copy(ws.yf.value)]
            self.debug["jacobian"]        += [np.copy(ws.jacobian.value)]
            self.debug["iteration_index"] += [ws.inversion_iteration_counter.value]

        if debug:
            agenda.append(get_debug)

        ws.inversion_iterate_agenda = agenda


    def run(self, *args, **kwargs):

        ws = self.simulation.workspace

        self.setup_iteration_agenda()
        self.setup_a_priori(*args, **kwargs)

        y_blocks = []
        for s in self.sensors:
            if isinstance(s, ActiveSensor):
                i, j = self.sensor_indices[s.name]
                y_blocks += [self.y[i : j]]
        for s in self.sensors:
            if not isinstance(s, ActiveSensor):
                i, j = self.sensor_indices[s.name]
                y_blocks += [self.y[i : j]]

        try:

            y = np.concatenate(y_blocks)
            self._y = y
            ws.y  = y
            ws.yf       = []
            ws.jacobian = []

            self.simulation.run_checks()
            ws.OEM(**self.settings)

        except Exception as e:
            ws.oem_diagnostics = 9 * np.ones(5)
            ws.yf       = None
            ws.jacobian = None
            ws.oem_errors = ["Error in OEM computation.", str(e)]

        self.x               = np.copy(ws.x.value)
        self.oem_diagnostics = np.copy(ws.oem_diagnostics)
        self.yf              = np.copy(ws.yf.value)
        self.jacobian        = np.copy(ws.jacobian.value)

        if self.oem_diagnostics[0] <= 2.0:
            ws.avkCalc()
            self.dxdy      = np.copy(ws.dxdy)
            self.jacobian  = np.copy(ws.jacobian)
            self.avk       = np.copy(ws.avk)
            ws.covmat_soCalc()
            self.covmat_so = np.copy(ws.covmat_so.value)
            ws.covmat_ssCalc()
            self.covmat_ss = np.copy(ws.covmat_ss.value)
        else:
            self.dxdy      = None
            self.jacobian  = None
            self.avk       = None
            self.covmat_so = None
            self.covmat_ss = None

        if self.oem_diagnostics[0] == 9.0:
            try:
                self.oem_errors = ws.oem_errors.value
                print("Error in OEM calculation:")
                print(self.oem_errors)
            except Exception as e:
                self.oem_errors = ["Error in OEM computation.", str(e)]

        else:
            self.oem_errors = None


class RetrievalCalculation:
    """
    The :class:`Retrieval` takes care of the book-keeping around retrieval
    quantities in an ARTS simulation as well as the execution of the retrieval
    calculation.

    Arguments:

        debug_mode(:code:`bool`): If set to true, debug information will be
            collected while the retrieval is run.
    """
    def __init__(self,
                 debug_mode = False):

        self.retrieval_quantities = []
        self.y = None

        self.settings = {"method" : "lm",
                         "max_start_cost" : np.inf,
                         "x_norm" : np.zeros(0),
                         "max_iter" : 20,
                         "stop_dx" : 0.1,
                         "lm_ga_settings" : np.array([1000.0, 5.0, 2.0, 1e6, 1.0, 1.0]),
                         "clear_matrices" : 0,
                         "display_progress" : 1}

        self.callbacks = []
        self.debug_mode = debug_mode

    def add(self, rq):
        """
        Add a retrieval quantity to the retrieval calculation.

        This registers an atmospheric quantity as a retrieval quantity,
        which means that artssat will try to retrieve the quantity using ARTS's
        OEM method instead of querying its value from the data provider. While
        the data provider is not required to provide get methods for
        the retrieval quantity itself, it must provide values for its a priori
        mean and covariance matrix.

        Arguments:

            rq(:code:`RetrievalQuantity`): The retrieval quantity for which to
                compute the retrieval.

        """
        rq.retrieval = rq.retrieval_class(rq, len(self.retrieval_quantities))
        self.retrieval_quantities += [rq]


    def _get_y_vector(self, simulation, *args, **kwargs):

        if len(simulation.sensors) == 0:
            raise Exception("Can't perform retrieval without sensors.")

        y = getattr(self, "y", None)
        if y is None:
            f = getattr(simulation.data_provider, "get_y", None)
            if not f is None:
                y = f(*args, **kwargs)
            else:
                ys = []

                for ss in [simulation.active_sensors, simulation.passive_sensors]:
                    for s in ss:
                        fname = "get_y_" + s.name
                        f = getattr(simulation.data_provider, fname, None)
                        if f is None:
                            raise Exception("No measurement vector provided for "
                                            " sensor {0}.".format(s.name))
                        ys += [f(*args, **kwargs).ravel()]
                y = np.concatenate(ys)
        return y

    def run(self, simulation, *args, **kwargs):
        """
        Setup the retrieval calculation.

        This methods performs the setup necessary to run the retrieval
        calculations on the given workspace. That means it does:

        1. Gather a priori mean states and covariance matrices for all
            registered retrieval quantities.

        2. Define the retrieval inside the given ARTS workspace.

        3. Construct the ARTS inversion iterate agenda.

        Arguments:

            ws(:code:`arts`): The :code:`pyarts.workspace.Workspace`
                object which to setup for the retrieval.

            sensors(list): The sensors to use for the retrieval calculation.

            scattering(:code:`bool`): Whether or not to the forward calculations
                involve scattering.

            retrieval_provider: The data provider providing the a priori data
                for the retrieval.

            *args, **kwargs: Arguments and keyword arguments that are passed on
                to the data provider.
        """

        # Determine sensor indices of y vector
        i_start = 0
        self.sensor_indices = {}
        for s in simulation.active_sensors:
            s.get_data(simulation.workspace, simulation.data_provider, *args, **kwargs)
            self.sensor_indices[s.name] = (i_start, i_start + s.y_vector_length)
            i_start += s.y_vector_length
        for s in simulation.passive_sensors:
            self.sensor_indices[s.name] = (i_start, i_start + s.y_vector_length)
            i_start += s.y_vector_length

        self._y = self._get_y_vector(simulation, *args, **kwargs)

        ws   = simulation.workspace

        previous_run = None
        if self.callbacks == []:
            #
            # No retrieval callback provided. The retrieval consists only
            # of a single run.
            #
            retrieval = RetrievalRun("Retrieval",
                                     simulation,
                                     self._y,
                                     self.settings,
                                     self.sensor_indices,
                                     self.retrieval_quantities)
            retrieval.run(*args, **kwargs)
            self.results = retrieval
        else:
            #
            # If a list of retrieval callbacks is provided, an OEM run
            # is run for each callback. Forwarding retrieval results to
            # the next call.
            #
            self.results = []
            for cb in self.callbacks:

                if type(cb) is tuple:
                    name, cb = cb
                else:
                    name = str(len(self.results))

                retrieval = RetrievalRun(name,
                                         simulation,
                                         self._y,
                                         self.settings,
                                         self.sensor_indices,
                                         self.retrieval_quantities,
                                         previous_run = previous_run)

                if not cb is None:
                    cb(retrieval)

                retrieval.run(*args, **kwargs)
                self.results += [retrieval]
                previous_run = retrieval



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='artssat',  # Required
    version='0.0.3',  # Required
    description='High-level simulation and retrieval framework for microwave and IR remote sensing.',
    long_description=long_description,
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://github.com/simonpf/artssat',  # Optional
    author='Simon Pfreundschuh',  # Optional
    author_email='simon.pfreundschuh@chalmers.se',  # Optional
    install_requires=["pyarts", "typhon", "netCDF4"],
    packages=find_packages(exclude=['examples', 'doc', 'misc', 'tests']),
    python_requires='>=3.6',
    project_urls={  # Optional
        'Source': 'https://github.com/simonpf/artssat/',
    })



""" ARTS Sensors

Class hierarchy that represents the different types of sensors
in ARTS. The purpose of this hierarchy is to encapsulate all
required settings and configurations related to the sensor
in an ARTS simulation.

An ARTS simulation contains one or more sensors. The interaction
between ARTS simulation and sensor object is structured as follows:

1. Setup: During the setup phase the ARTS simulation calls the
   setup method of each sensor object. During this call the sensor
   is expected to setup all necessary workspace variables on
   the provided workspace and perform all necessary preparatory
   calculations, such as calculating absorption and scattering
   data.

2. Run: During the run phase, the simulation calls the get_data(...)
   method of the sensor during which the sensor is supposed to
   request all necessary data, that hasn't been set in advance.

In addition to that, the simulation needs to know which WSMs
to call to compute the measurement vector. To this end, each
sensor needs to provide the following factory methods:

1. make_preparation_function: This function should generate
   a function that runs all necessary preparations on a
   provided workspace. Currently, this requires setting
   only the stokes dimension (because of an inconsistency in ARTS).

2. make_y_calc_function: This function should generate a
   functions that runs the necessary workspace methods
   to simulate the measurements on a given workspace.

Both of these factory methods must produce functions that
can be converted into an ARTS agenda, so that they can be
used also within the inversion iterate agenda for retrievals.

In general, sensors should store important within their own,
anonymous workspace variables, which should then just be
used to replace the input WSV in all sensor-related WSM
calls. Due to the inconsistency metioned abover, however, this
principle does not work for the stokes dimension as it is
not consistently passed to the surface agenda.

Attributes:

    wsm(dict): Alias for the :code:'workspace_methods' dictionary
               from 'pyarts.workspace.methods'
    wsv(dict): Alias for the :code:`workspace_variables` dictionary
               from :code:`pyarts.workspace.variables`
"""
from abc import ABCMeta, abstractmethod, abstractproperty

import numpy as np

from artssat.arts_object import ArtsObject, arts_property
from artssat.arts_object import Dimension as dim
from pyarts.types import SingleScatteringData
from pyarts.workspace import Workspace, arts_agenda
from pyarts.workspace.methods import workspace_methods
from pyarts.workspace.variables import WorkspaceVariable, \
                                            workspace_variables

wsv = workspace_variables
wsm = workspace_methods

################################################################################
### Abstract sensor class
################################################################################

class Sensor(ArtsObject):
    """
    Defines an interface and implementes common functionality for
    classes representing ARTS sensors.

    Declares ARTS properties common to all types of sensors handled
    by ARTS.

    Arts Attributes:

        properties: The ARTS properties that hold the data common to
            all sensors in ARTS.
    """

    private_wsvs = ["f_grid", "scat_data", "scat_data_checked",
                    "iy_unit", "iy_aux_vars", "sensor_los", "sensor_pos",
                    "sensor_time", "mblock_dlos_grid", "sensor_response",
                    "antenna_dim", "sensor_norm", "sensor_response",
                    "sensor_response_f", "sensor_response_f_grid",
                    "sensor_response_dlos", "sensor_response_dlos_grid",
                    "sensor_response_pol", "sensor_response_pol_grid",
                    "iy_main_agenda"]


    ############################################################################
    # ARTS properties
    ############################################################################

    @arts_property("Vector",
                   shape = (dim.Joker,),
                   wsv = wsv["f_grid"])
    def f_grid(self):
        """
        The frequency grid of the sensor.
        """
        return None

    @arts_property("String",
                   wsv = wsv["iy_unit"])
    def iy_unit(self):
        """
        The unit which to use for the measurement vector.
        """
        return "1"

    @arts_property("ArrayOfString",
                   shape = (dim.Joker,),
                   wsv = wsv["iy_aux_vars"])
    def iy_aux_vars(self):
        """
        Which auxilliary variables to compute dunring the simulation.
        """
        return []

    @arts_property("Integer", wsv = wsv["stokes_dim"])
    def stokes_dimension(self):
        """
        Which stokes dimensions to use for the simulations.
        """
        return 1

    @arts_property("Matrix",
                   shape = (dim.Joker, dim.Atm),
                   wsv = wsv["sensor_pos"])
    def sensor_position(self):
        """
        For which sensor positions ot perform the simulations.
        """
        return []

    @arts_property("Matrix",
                   shape = (dim.Joker, dim.Atm),
                   wsv = wsv["transmitter_pos"])
    def transmitter_pos(self):
        """
        Positions of the transmitter for radio link calculations.
        """
        return []

    @arts_property("Matrix",
                   shape = (dim.Obs, dim.Los),
                   wsv = wsv["sensor_los"])
    def sensor_line_of_sight(self):
        """
        Line of sight for each sensor position.
        """
        return []

    @arts_property("Matrix",
                   shape = (dim.Obs, dim.Los),
                   wsv = wsv["mblock_dlos_grid"])
    def sensor_line_of_sight_offsets(self):
        """
        Line of sight offsets for each measurment block.
        """
        return np.zeros((1, self.sensor_line_of_sight.shape[1]))

    @arts_property("Sparse",
                   shape = (dim.Joker, dim.Joker),
                   wsv = wsv["sensor_response"])
    def sensor_response(self):
        """
        The sensor response matrix.
        """
        return []

    @arts_property("Vector",
                   shape = (dim.Joker,),
                   wsv = wsv["sensor_response_f"])
    def sensor_response_f(self):
        """
        The frequencies associated with the rows of the sensor response matrix.
        """
        return []

    @arts_property("ArrayOfIndex",
                   shape = (dim.Joker,),
                   wsv = wsv["sensor_response_pol"])
    def sensor_response_pol(self):
        """
        The polarization states associated with the rows of the sensor reponse
        matrix.
        """
        return []

    @arts_property("Vector",
                   shape = (dim.Joker, dim.Joker),
                   wsv = wsv["sensor_response_dlos"])
    def sensor_response_dlos(self):
        """
        The line-of-sight offsets associates with the rows of the sensor
        response matrix.
        """
        return []

    @arts_property("Vector",
                   shape = (dim.Joker,),
                   wsv = wsv["sensor_response_f_grid"])
    def sensor_response_f_grid(self):
        """
        The frequency grid associated with sensor response.
        """
        return []

    @arts_property("ArrayOfInteger",
                   shape = (dim.Joker,),
                   wsv = wsv["sensor_response_pol_grid"])
    def sensor_response_pol_grid(self):
        """
        The polarization states associated with sensor response.
        """
        return []

    @arts_property("Vector",
                   shape = (dim.Joker,),
                   wsv = wsv["sensor_response_dlos_grid"])
    def sensor_response_dlos_grid(self):
        """
        The LOS offsets associated with sensor response.
        """
        return []

    @arts_property("Vector",
                   shape = (dim.Joker,),
                   wsv = wsv["sensor_time"])
    def sensor_time(self):
        """
        The times associated with each measurment block.
        """
        return np.array(np.zeros(self.sensor_position.shape[0]))

    @arts_property("Index",
                   wsv = wsv["sensor_norm"])
    def sensor_norm(self):
        """
        Flag whether or not to normalize the sensor response.
        """
        return 1
    @arts_property("Index",
                   wsv = wsv["antenna_dim"])
    def antenna_dim(self):
        """
        The dimensionality of the antenna pattern.
        """
        return 1

    ############################################################################
    # Class methods
    ############################################################################

    def __init__(self, name, f_grid = None, stokes_dimension = 1):
        """
        Create a sensor with given frequency grid :code:`f_grid` and
        stokes dimension :code: `stokes_dimension`
        """
        super().__init__()
        self._wsvs = {}

        if not f_grid is None:
            self.f_grid = f_grid

        self.name = name
        self.stokes_dimension = stokes_dimension

    #
    # Abstract properties and methods
    #

    @abstractmethod
    def make_iy_main_agenda(self, scattering = False):
        """
        Property that takes the role of a factory function that creates
        the :code:`iy_main_agenda` for the given sensor. This will
        depend on the sensor type and its specific settings, so
        must be implemented by the inhereting classes.
        """
        pass

    @abstractmethod
    def make_preparation_function(self, append = False):
        """
        Factory method for a preparation function :code:`f(ws)` which
        sets all workspace variables that are required before simulating
        a measurement on a given workspace :code:`ws`.

        The separation into a preparation function and a :code:`y_calc`
        function is currently necessary as the scattering solver can
        be chosen independently from the sensor, so the simulation object
        must be able to run the scattering solver between the sensor
        preparations and running :code:`y_calc`.

        Parameters:
            append(bool): Whether the call should use :code:`yCalcAppend`
                or not. Active sensors will probably have to throw an
                exception here.
        """
        pass

    @abstractmethod
    def make_y_calc_function(self,
                             append = False,
                             scattering = False):
        """
        Factory method that should create a function :code:`f` that runs
        the actual radiative transfer on a provided workspace.

        Parameters:
            append(bool): Whether the call should use :code:`yCalcAppend`
                or not. Active sensors will probably have to throw an
                exception here.
        """
        pass

    #
    # Special setters
    #

    @stokes_dimension.setter
    def stokes_dimension_setter(self, n):
        """
        Specialized setter that overwrites the default from the
        :code:`ArtsObject` base class. Makes sure that the stokes
        dimension is 1,2 or 4.

        Parameters:

            n(int): The new stokes dimension

        Raises:

            Exception if the new value for the stokes dimension is
            not 1, 2 or 4.
        """
        if not (n in [1, 2, 4]):
            raise Exception("Stokes dimension must be 1, 2 or 4.")
        self._stokes_dimension.fixed = True
        self._stokes_dimension.value = n

    @abstractproperty
    def y_vector_length(self):
        pass

    @property
    def views(self):
        return self.sensor_position.shape[0]

    #
    # General sensor setup
    #

    def setup(self, ws, scattering = True):
        """
        General setup for an ARTS sensor.

        This method performs the following steps to setup the sensor:

        - Copy the `iy_main_agenda` of the sensor into a private workspace
          variable in the workspace
        - Copy the `f_grid` into a private workspace variable
        - Compute and check `scat_data` and copy results into private
          workspace variables.
        - Copy `iy_aux_vars` and `iy_unit` to the workspace and store
          in prive workspace variables.

        Paremters:

            ws(pyarts.workspace.Workspace): The workspace on which
                to perform the setup of the sensor.
        """
        self._create_private_wsvs(ws, type(self).private_wsvs)
        wsvs = self._wsvs

        self.setup_arts_properties(ws)

        #
        # Scat data
        #
        if ws.scat_data_raw.initialized:
            ws.scat_dataCalc(scat_data_raw=ws.scat_data_raw,
                             f_grid=wsvs["f_grid"],
                             interp_order=1)
            ws.Copy(wsvs["scat_data"], ws.scat_data)

        kwargs = self.get_wsm_kwargs(wsm["scat_data_checkedCalc"])
        ws.scat_data_checkedCalc(**kwargs, check_level = "sane")
        wsvs["scat_data_checked"].value = ws.scat_data_checked.value

        #
        # Need to add agendas in the end so that input arguments
        # can be replaced by private sensor variables.
        #
        wsvs["iy_main_agenda"].value = self.make_iy_main_agenda(scattering)

    def get_data(self, ws, data_provider, *args, **kwargs):
        """
        Get required data from data provided.

        This function obtains required data from the data provider
        if it has not been fixed in advance. The data expected
        from the data provider are the following:

        - the line of sight data by calling the :code:`get_line_of_sight`
          method of the provided data provider.
        - the sensor position data by calling the :code: `get_sensor_pos`
          method of the provided data provider.
        - the pencil beam offsets by calling the :code:
          `get_line_of_sight_offsets`

        This function also checks the sensor data by calling
        :code:`sensor_checkedCalc` and stores the result in a private
        WSV.

        """
        self.get_data_arts_properties(ws, data_provider, *args, **kwargs)
        if isinstance(self.sensor_response, list) and self.sensor_response == []:
            self.call_wsm(ws, wsm["sensor_responseInit"])
        self.call_wsm(ws, wsm["sensor_checkedCalc"])



################################################################################
### Active sensor class
################################################################################

class ActiveSensor(Sensor):
    """
    Specialization of the abstract :code:`Sensor` class that implements
    active sensors (Radar).

    """
    ws = Workspace()
    extinction_scaling = ws.create_variable("Numeric", "extinction_scaling")
    private_wsvs = Sensor.private_wsvs + ["range_bins",
                                          "instrument_pol_array",
                                          "instrument_pol",
                                          "iy_transmitter_agenda",
                                          "extinction_scaling"]

    ############################################################################
    # ARTS properties
    ############################################################################

    @arts_property("Numeric", wsv = "extinction_scaling")
    def extinction_scaling(self):
        return 1.0

    @arts_property("Numeric")
    def y_min(self):
            return -35.0

    @arts_property("Vector", shape = (dim.Joker,), wsv = wsv["range_bins"])
    def range_bins(self):
        return []

    @arts_property("ArrayOfIndex",
                   wsv = wsv["instrument_pol"])
    def instrument_pol(self):
        return [1]

    @arts_property("ArrayOfArrayOfIndex",
                   wsv = wsv["instrument_pol_array"])
    def instrument_pol_array(self):
        return [[1]]

    @property
    def y_vector_length(self):
        return (self.range_bins.size - 1) * self.f_grid.size * self.stokes_dimension

    def __init__(self, name, f_grid, stokes_dimension, range_bins = None):
        super().__init__(name, f_grid, stokes_dimension = stokes_dimension)
        self.iy_unit    = "dBZe"

        if not range_bins is None:
            self.range_bins = range_bins

    #
    # Agendas
    #

    @property
    def iy_transmitter_agenda(self):
        """
        The :code:`iy_transmitter_agenda` which is required for active
        sensors. Input arguments of :code:`iy_transmitter_agenda` are
        replaced by the private workspace variables of the sensor.

        Returns:

            The iy_transmitter_agenda for the active sensor.

        """
        kwargs = self.get_wsm_kwargs(wsm["iy_transmitterSinglePol"])
        @arts_agenda
        def iy_transmitter_agenda(ws):
            ws.Ignore(ws.rtp_pos)
            ws.Ignore(ws.rtp_los)
            ws.Ignore(ws.f_grid)
            ws.iy_transmitterSinglePol(**kwargs)

        return iy_transmitter_agenda

    def make_iy_main_agenda(self, scattering = False):
        """
        The :code: `iy_main_agenda` for active sensor. Currently uses
        the single scattering radar module, but might be extended
        at some point.
        """

        kwargs = self.get_wsm_kwargs(wsm["iyActiveSingleScat2"])
        @arts_agenda
        def iy_main_agenda(ws):
            ws.Ignore(ws.iy_id)
            ws.Ignore(ws.nlte_field)
            ws.Ignore(ws.rte_pos2)
            ws.Ignore(ws.iy_unit)
            ws.Ignore(ws.iy_aux_vars)
            ws.FlagOff(ws.cloudbox_on)
            ws.ppathCalc()
            ws.FlagOn(ws.cloudbox_on)
            ws.iyActiveSingleScat(**kwargs,
                                  pext_scaling = self._wsvs["extinction_scaling"],
                                  trans_in_jacobian = 1)
        return iy_main_agenda

    #
    # Specialized setters
    #

    def iy_unit_setter(self, u):
        if not u in ["1", "Ze", "dBZe"]:
            raise Exception("Value of iy_unit for an active sensor must"
                            " be one of ['1', 'Ze', 'dBZe']")
        else:
            self._iy_unit.value = u
            self._iy_unit.fixed = True

    def iy_aux_vars_setter(self, v):
        if not type(v) == list:
            v = [v]
        if not all([u in ["Radiative background", "Backsacttering",
                         "Optical depth", "Particle extinction"] for u in v]):
            raise Exception("Value of iy_aux_vars for an active sensor must"
                            " be a list consisting of the following strings: "
                            "['RadiativeBackground', 'Backscattering', "
                            "'Optical depth', Particle extinction'].")
        else:
            self._iy_aux_vars.value = v
            self._iy_aux_vars.fixed = False

    #
    # Preparation and y_calc factories.
    #

    def make_preparation_function(self):
        """
        Return the workspace preparation function, which prepares
        a workspace for simulating the signal recorded by the sensor.

        Returns:

            The function to prepare the workspace.

        """

        def preparations(ws):
            ws.IndexSet(ws.stokes_dim, self.stokes_dimension)
            #ws.Copy(ws.iy_transmitter_agenda,
            #        self._wsvs["iy_transmitter_agenda"])
            #ws.Copy(ws.iy_main_agenda,
            #        self._wsvs["_iy_main_agenda"])
            #ws.Copy(ws.instrument_pol, self._wsvs["instrument_pol"])
            #ws.IndexSet(self._wsvs["_stokes_dim"], self.stokes_dimension)
            #ws.IndexSet(ws.stokes_dim, self.stokes_dimension)

        return preparations

    def make_y_calc_function(self,
                             append = False,
                             scattering = False):
        """
        Returns y_calc function, which computes the radar signal
        on an accordingly prepared workspace. This function can
        be converted into an ARTS agenda and thus included in
        the other agendas using the INCLUDE statement.

        Returns:

            The function to compute the radar signal.
        """
        if append:
            raise Exception("ARTS doesn't support appending measurements from"
                            " active sensors.")

        kwargs = self.get_wsm_kwargs(wsm["yActive"])

        if self.y_min:
            y_min = self.y_min
        else:
            y_min = - np.inf

        def y_calc(ws):
            ws.yActive(**kwargs, dbze_min = y_min)

        return y_calc

    #
    # Setup
    #

    def setup(self, ws, scattering = True):
        super().setup(ws, scattering)
        self._wsvs["iy_transmitter_agenda"].value = self.iy_transmitter_agenda

class PassiveSensor(Sensor):
    """
    Specialization of the abstract Sensor class for passive sensors.
    """

    @arts_property("Index")
    def t_interp_order(self):
        return 0

    @property
    def y_vector_length(self):
        if hasattr(self.sensor_response, "shape"):
            return self.sensor_response.shape[0]
        else:
            return self.f_grid.size * self.stokes_dimension

    def __init__(self, name, f_grid, stokes_dimension = 1):
        """
        Paramters:
            f_grid(numpy.ndarray) The frequency grid of the sensor.
            stokes_dimension(int) The stokes dimensions to use for simulating
                the sensor measurement. Must be one of [1, 2, 4].
        """
        super().__init__(name, f_grid, stokes_dimension)
        self.iy_unit = "PlanckBT"

    #
    # Agendas
    #

    def make_iy_main_agenda(self, scattering = False):
        """
        Factory property that returns the :code:`iy_main_agenda` that has to be used
        for passive sensors.

        Return:

            The ARTS :code:`iy_main_agenda`
        """
        def iy_main_agenda_scattering(ws):
            ws.Ignore(ws.iy_id)
            ws.Ignore(ws.nlte_field)
            ws.Ignore(ws.rte_pos2)
            ws.Ignore(ws.iy_unit)
            ws.Ignore(ws.iy_aux_vars)
            ws.FlagOff(ws.cloudbox_on)
            ws.ppathCalc()
            ws.FlagOn(ws.cloudbox_on)
            ws.iyHybrid2(**kwargs, t_interp_order = self.t_interp_order)

        def iy_main_agenda_no_scattering(ws):
            ws.Ignore(ws.iy_id)
            ws.Ignore(ws.nlte_field)
            ws.Ignore(ws.rte_pos2)
            ws.Ignore(ws.iy_unit)
            ws.Ignore(ws.iy_aux_vars)
            ws.ppathCalc()
            ws.iyEmissionStandard(**kwargs)

        if scattering:
            agenda = iy_main_agenda_scattering
            kwargs = self.get_wsm_kwargs(wsm["iyHybrid2"])
        else:
            agenda = iy_main_agenda_no_scattering
            kwargs = self.get_wsm_kwargs(wsm["iyEmissionStandard"])

        agenda.__name__ = "iy_main_agenda"

        return arts_agenda(agenda)

    #
    # Customized setter
    #

    def iy_unit_setter(self, u):
        """
        Custom setter for :code:`iy_unit` for passive sensors. Checks
        that the argument :code:`u` has a valid value for passive simulations.

        Parameters:
            u(str): The :code:`iy_unit` to use for passive simulations. Must
                be one of :code:`["1", "RJBT", "PlackBT", "W(m^2 m sr)",
                "W/(m^2 m-1 sr)"]`

        """
        valid = ["1", "RJBT", "PlanckBT", "W/(m^2 m sr)", "W/(m^2 m-1 sr)"]
        if not u in valid:
            raise Exception("Value of iy_unit for an active sensor must"
                            " be one of {0}".format(str(valid)))
        else:
            self._iy_unit = u

    def iy_aux_vars_setter(self, v):
        """
        Custom setter for :code:`iy_aux_vars` for passive sensors. Checks
        that the argument :code:`v` has a valid value for passive simulations.

        Parameters:
            u(str): The :code:`iy_aux_vars` to use for passive simulations. Must
                be a list containing any of :code:`["Radiative background",
                "Optical depth]`

        """
        if not type(v) == list:
            v = [v]
        if not all([u in ["Radiative background", "Optical depth"]
                    for u in v]):
            raise Exception("Value of iy_aux_vars for an active sensor must"
                            " be a list consisting of the following strings: "
                            "['RadiativeBackground', 'Optical depth'].")
        else:
            self._iy_aux_vars = v

    #
    # Preparation and y_calc factory methods
    #

    def make_preparation_function(self):
        """
        Return the workspace preparation function, which prepares
        a workspace for simulating the signal recorded by the sensor.

        Returns:

            The function to prepare the workspace.

        """

        def preparations(ws):
            ws.IndexSet(ws.stokes_dim, self.stokes_dimension)

        return preparations

    def make_y_calc_function(self,
                             append = False,
                             scattering = False):
        """
        Factory function that produces a function that simulates a passive
        measurement with the given sensor on an accordingly prepared workspace.

        If :code:`append` is :code:`False` the returned function will use
        the :code:`yCalc` WSV to simulate the measurement. Otherwise
        :code:`yCalcAppend` will be used.

        Parameters:

            append(bool): If :code:`True` the produced function will use
                `yCalcAppend` to calculate the measured signal.
        """

        if append:
            kwargs = self.get_wsm_kwargs(wsm["yCalcAppend"])
        else:
            kwargs = self.get_wsm_kwargs(wsm["yCalc"])

        def y_calc_append(ws):
            ws.yCalcAppend(**kwargs,
                           jacobian_quantities_copy = ws.jacobian_quantities)

        def y_calc(ws):
            ws.yCalc(**kwargs)

        if append:
            f = y_calc_append
        else:
            f = y_calc

        return f


################################################################################
# Ice cloud imager (ICI).
################################################################################

class ICI(PassiveSensor):
    """
    The Ice Cloud Imager (ICI) sensor.

    Attributes:

        channels(:code:`list`): List of channels that are available
            from ICI

        nedt(:code:`list`): Noise equivalent temperature differences for the
            channels in :code:`channels`.
    """
    channels = np.array([1.749100000000000e+11,
                         1.799100000000000e+11,
                         1.813100000000000e+11,
                         2.407000000000000e+11,
                         3.156500000000000e+11,
                         3.216500000000000e+11,
                         3.236500000000000e+11,
                         4.408000000000000e+11,
                         4.450000000000000e+11,
                         4.466000000000000e+11,
                         6.598000000000000e+11])

    nedt = np.array([0.8, 0.8, 0.8,       # 183 GHz
                     0.7 * np.sqrt(0.5),  # 243 GHz
                     1.2, 1.3, 1.5,       # 325 GHz
                     1.4, 1.6, 2.0,       # 448 GHz
                     1.6 * np.sqrt(0.5)]) # 664 GHz

    def __init__(self,
                 name = "ici",
                 channel_indices = None,
                 stokes_dimension = 1,
                 lines_of_sight = None,
                 positions = None):
        """
        This creates an instance of the ICI sensor to be used within a
        :code:`artssat` simulation.

        Arguments:

            name(:code:`str`): The name of the sensor used within the artssat
                simulation.

            channel_indices(:code:`list`): List of channel indices to be used
                in the simulation/retrieval.

            stokes_dimension(:code:`int`): The stokes dimension to use for
                the retrievals.
        """
        if channel_indices is None:
            channels  = ICI.channels
            self.nedt = ICI.nedt
        else:
            channels  = ICI.channels[channel_indices]
            self.nedt = self.nedt[channel_indices]
        super().__init__(name, channels, stokes_dimension = stokes_dimension)

        if not (lines_of_sight is None):
            if not (positions is None):
                self.sensor_line_of_sight = lines_of_sight
                self.sensor_position = positions
            else:
                self.sensor_position = np.array([[600e3]])
                self.sensor_line_of_sight = np.array([[135.0]])
        else:
            self.sensor_position = np.array([[600e3]])
            self.sensor_line_of_sight = np.array([[135.0]])

################################################################################
# Microwave imager (MWI).
################################################################################

class MWI(PassiveSensor):
    """
    The Microwave Imager (MWI) sensor.

    Attributes:

        channels(:code:`list`): The list of the channels available from the
            MWI sensor.

        nedt(:code:`list`): The noise equivalent temperature differences for
            the channels in :code:`channels`.
    """
    channels = np.array([18.7e9,
                         23.8e9,
                         31.4e9,
                         50.3e9,
                         52.6e9,
                         53.24e9,
                         53.75e9,
                         89.0e9,
                         115.5503e9,
                         116.6503e9,
                         117.3503e9,
                         117.5503e9,
                         164.75e9,
                         176.31e9,
                         177.21e9,
                         178.41e9,
                         179.91e9,
                         182.01e9])

    nedt = np.array([0.8 * np.sqrt(0.5), #18 GHz
                     0.7 * np.sqrt(0.5), #24 GHz
                     0.9 * np.sqrt(0.5), #31 GHz
                     1.1 * np.sqrt(0.5), #50 GHz
                     1.1 * np.sqrt(0.5),
                     1.1 * np.sqrt(0.5),
                     1.1 * np.sqrt(0.5),
                     1.1 * np.sqrt(0.5), #89 GHz
                     1.3, #118 GHz
                     1.3,
                     1.3,
                     1.3,
                     1.2, #165 GHz
                     1.3, #183 GHz
                     1.2,
                     1.2,
                     1.2,
                     1.3])

    def __init__(self,
                 name = "mwi",
                 channel_indices = None,
                 stokes_dimension = 1):
        """
        Create an MWI instance to be used within a :code:`artssat` simulation.

        Arguments:

            name(:code:`str`): The name of the sensor to be used within the
                artssat simulation.

            channel_indices(:code:`list`): List of channel indices to be used
                for the simulation.

            stokes_dimension(:code:`int`): The Stokes dimension to be used for
                the simulation.
        """
        if channel_indices is None:
            channels  = MWI.channels
            self.nedt = MWI.nedt
        else:
            channels  = MWI.channels[channel_indices]
            self.nedt = MWI.nedt[channel_indices]

        self.channels = channels
        super().__init__(name, channels, stokes_dimension = stokes_dimension)

class CloudSat(ActiveSensor):
    channels = np.array([94.0e9])

    def __init__(self,
                 name = "cloud_sat",
                 range_bins = np.arange(500.0, 20e3, 500.0),
                 stokes_dimension = 2,
                 lines_of_sight = None,
                 positions = None):
        super().__init__(name,
                         f_grid = np.array([94e9]),
                         stokes_dimension = stokes_dimension,
                         range_bins = range_bins)
        self.instrument_pol       = [1]
        self.instrument_pol_array = [[1]]
        self.extinction_scaling   = 1.0
        self.y_min = -30.0

        if not (lines_of_sight is None):
            if not (positions is None):
                self.sensor_line_of_sight = lines_of_sight
                self.sensor_position = positions
            else:
                self.sensor_position = np.array([[600e3]])
                self.sensor_line_of_sight = np.array([[180.0]])
        else:
            self.sensor_position = np.array([[600e3]])
            self.sensor_line_of_sight = np.array([[180.0]])

ici = ICI()



"""

The :code:`artssat` package provides a modular framework for
radiative transfer simulations. It does so by defining a general
computing workflow that allows hiding away much of the complexity
of ARTS behind physically-motivated Python classes.

Overview
--------

The :class:`ArtsSimulation` class provides the main user interface
of the :code:`artssat` package. As the name suggests, it represents
an ARTS simulation, which may be a simple forward simulation, a
Jacobian calculation or a retrieval calculation.

The :code:`atmosphere` attribute of an :class:`ArtsSimulation` object
represents the model atmosphere that is underlying the radiative transfer
simulation. What kind of radiative transfer simulations are performed on the
model atmosphere is defined by the :code:`sensors` attribute of the simulation.
These are the main components that describe the general simulation *setup*.

It is important to note that a :class:`ArtsSimulation` object represents only
an abstract description of the simulation scenario. That means it doesn't hold
any data describing a specific atmospheric state, such as for example the
temperature field. The rationale for this design decision is to separate the
simulation setup from the data source, so that these two can be exchanged more
easily.

The data describing a concrete atmsopheric state to be simulated is expected
to be provided by a **data provider**. The data providers task is to provide
get methods for all data required for the simulation. These get methods may
be parametrized to allow the simulation of several different states. When
the :code:`run(*args, **kwargs)` method of the :class:`ArtsSimulation` is
called, it forwards its arguments to the getter methods of the data provider.
For a more detailed description of the data flow see the section below.

The image below illustrates the general semantic model used to describe a
radiative transfer simulation.

.. image:: /figures/overview_diagram.svg

Interaction and data workflow
-----------------------------

The general interaction and data flow for a single simulation is illustrated
in the figure below.

After defining the simulation setup, preparatory calculations are performed
by calling the :code:`setup()` method of the :class:`ArtsSimulation` object.
To then run a simulation, the user calls the :code:`run(...)` method providing
as arguments the arguments that should be passed on to the :code:`data_provider`
to return the data of the atmospheric state to simulate (Step 1).

Depending on the simulation setup, the :class:`ArtsSimulation` object will
require different data to perform the simulation. These quantities
are identified by a unique name :code:`name` and for each of these the
:class:`ArtsSimulation` object will call the corresponding :code:`get_<name>`
method of the :code:`data_provider` with the arguments provided by the
user to the :code:`run` method (Step 2).

After receiving the required data from the :code:`data_provider` (Step 3),
the :class:`ArtsSimulation` object performs the simulation and returns the
result to the user (Step 4).

.. image:: /figures/data_flow_diagram.svg

"""
import artssat.dimensions
from artssat.simulation import ArtsSimulation



"""
artssat.models
==============

This module contains preconfigured :class:`artssat.ArtsSimulation` objects
that provide a starting point for performing radiative transfer calculations.
"""
from artssat.atmosphere.atmosphere import Atmosphere
from artssat.atmosphere.absorption import O2, N2, H2O
from artssat.atmosphere.surface    import Tessem

class StandardAtmosphere(Atmosphere):
    def __init__(self,
                 dimensions = 1,
                 surface = "ocean"):

        if not dimensions in [1]:
            raise Exception("Currently only 1D simulations are supported.")

        if not surface == "ocean":
            raise Exception("Currently only simulations over ocean surfaces are supported.")


        super().__init__(dimensions = (0, ) * dimensions,
                         absorbers = [O2(), N2(), H2O()],
                         scatterers = [],
                         surface = Tessem())






"""
Absorption
==========

"""
from abc import ABCMeta, abstractmethod, abstractproperty
from artssat.atmosphere.atmospheric_quantity \
    import AtmosphericQuantity, extend_dimensions

from artssat.arts_object import ArtsObject, arts_property
from artssat.jacobian    import JacobianBase
from artssat.retrieval   import RetrievalBase, RetrievalQuantity

import numpy as np
from pyarts.workspace import arts_agenda
from typhon.physics.atmosphere import vmr2relative_humidity, \
    relative_humidity2vmr

################################################################################
# Retrieval units
################################################################################

class Unit(metaclass = ABCMeta):
    """
    Abstract base class for classes representing units used for the calculation
    of Jacobians and retrievals of absorption species.
    """
    def __init__():
        pass

    @abstractmethod
    def to_arts(self, ws, x):
        pass

    @abstractmethod
    def from_arts(self, ws, x):
        pass

    @abstractproperty
    def arts_name(self):
        pass


class VMR(Unit):
    """
    VMR is the default unit used for absorption species in ARTS. If this unit
    is used value from the state vector are plugged in as they are into
    the ARTS vmr field.
    """

    def __init__(self):
        pass

    def to_arts(self, ws, x):
        """
        Does nothing.
        """
        return x

    def from_arts(self, ws, x):
        """
        Does nothing.
        """
        return x

    @property
    def arts_name(self):
        return "vmr"

class Relative(Unit):
    """
    In relative units, the amount of a quantity is specified relative to a
    reference profile or field. 

    If this unit is used in a Jacobian calculation, then the Jacobian is
    calculated w.r.t. a relative perturbation.

    If this unit is used in the retrieval, values in the state vector are
    interpreted as multiplicative perturbations of the reference profile
    or field.
    """
    def __init__(self, x_ref):
        self.x_ref = x_ref

    def to_arts(self, ws, x):
        return self.x_ref * x

    def from_arts(self, ws, y):
        return y / self.x_ref

    @property
    def arts_name(self):
        return "rel"

class RelativeHumidity(Unit):
    """
    Relative humidity is available only for the retrieval of H2O.
    """
    def __init__(self):
        pass

    def to_arts(self, ws, rh):
        """
        Converts value given in relative humidity units to ARTS vmr units.

        Arguments:

            ws(:code:`pyarts.workspace.Workspace`): Workspace object
                which contains pressure grid and temperature field required
                for the converstion.

            rh(:code:`numpy.ndarray`): Relative humidity values to convert.

        Returns:

            :code:`numpy.ndarray` containing the converted RH values.

        """
        p   = ws.p_grid.value.reshape(-1, 1, 1)
        t   = ws.t_field.value
        vmr = relative_humidity2vmr(rh, p, t)
        return vmr

    def from_arts(self, ws, vmr):
        """
        Converts a value given in ARTS vmr units back to relative humidity.

        Arguments:

            ws(:code:`pyarts.workspace.Workspace`): Workspace object
                which contains pressure grid and temperature field required
                for the conversion.

            vmr(:code:`numpy.ndarray`): Values in ARTS vmr units to convert
                to relative humidity.

        Returns:

            :code:`numpy.ndarray` containing the converted RH values.

        """
        p  = ws.p_grid.value.reshape(-1, 1, 1)
        t  = ws.t_field.value
        rh = vmr2relative_humidity(vmr, p, t)
        return rh

    @property
    def arts_name(self):
        return "rh"

################################################################################
# The Jacobian class
################################################################################

class Jacobian(JacobianBase, ArtsObject):

    @arts_property("Numeric")
    def perturbation(self):
        return 0.01

    def __init__(self, quantity, index):

        JacobianBase.__init__(self, quantity, index)
        ArtsObject.__init__(self)

        self.unit = VMR()
        self.for_species_tag = 1

    def _make_setup_kwargs(self, ws):
        kwargs = self.get_grids(ws)
        kwargs.update({"species" : self.quantity.get_tag_string(),
                       "unit" : self.unit.arts_name,
                       "for_species_tag" : self.for_species_tag})
        return kwargs

    def setup(self, ws):
        kwargs = self._make_setup_kwargs(ws)
        ws.jacobianAddAbsSpecies(**kwargs)

class Retrieval(RetrievalBase, Jacobian):

    def __init__(self, quantity, index):
        super().__init__(quantity, index)

    def add(self, ws):
        ws.retrievalAddAbsSpecies(**self._make_setup_kwargs(ws))


class AbsorptionSpecies(AtmosphericQuantity, RetrievalQuantity):

    def __init__(self,
                 name,
                 from_catalog = False,
                 cia = None,
                 frequency_range = None,
                 isotopologues = None,
                 model = None,
                 on_the_fly = True,
                 zeeman = False,
                 lineshape = "no_shape",
                 normalization = "no_norm",
                 cutoff = -1,
                 cutoff_type = "ByBand"):


        AtmosphericQuantity.__init__(self,
                                     name,
                                     (0, 0, 0))
        RetrievalQuantity.__init__(self)

        self._dimensions = (0, 0, 0)

        self._from_catalog    = from_catalog
        self._cia             = cia
        self._frequency_range = frequency_range
        self._isotopologues   = isotopologues
        self._jacobian        = None
        self._model           = model
        self._on_the_fly      = on_the_fly
        self._retrieval       = None
        self._zeeman          = zeeman

        self._lineshape = lineshape
        self._normalization = normalization
        self._cutoff = cutoff
        self._cutoff_type = cutoff_type


    #
    # Abstract properties
    #

    def dimensions(self):
        return self._dimensions

    #
    # Properties
    #

    @property
    def from_catalog(self):
        return self._from_catalog

    @property
    def cia(self):
        return self._cia

    @property
    def isotopologues(self):
        return self._isotopologues

    @property
    def model(self):
        return self._model

    @property
    def frequency_range(self):
        return self._frequency_range

    @property
    def on_the_fly(self):
        return self._on_the_fly

    @property
    def zeeman(self):
        return self._zeeman

    @property
    def lineshape(self):
        return self._lineshape

    @property
    def cutoff(self):
        return self._cutoff

    @property
    def cutoff_type(self):
        return self._cutoff_type

    @property
    def normalization(self):
        return self._normalization

    def _get_tag_string(self,
                        zeeman = False,
                        isotopologue = None,
                        model = None,
                        frequency_range = None):

        ts = self._name
        ts += "-"

        if zeeman:
            ts += "Z"
            ts += "-"

        if not isotopologue is None:
            ts += isotopologues
            ts += "-"

        if not model is None:
            ts += model
            ts += "-"

        if not frequency_range is None:
           ts += frequency_range[0]
           ts += "-"
           ts += frequency_range[1]
        return ts

    def get_tag_string(self):

        tss = []

        if self.from_catalog:
            tss += [self._name]

        z = self._zeeman

        if type(self._isotopologues) is list:
            isotopologues = self._isotopologues
        else:
            isotopologues = [self._isotopologues]

        if type(self._model) is list:
            models = self._model
        else:
            models = [self._model]

        if type(self._frequency_range) is list:
            frequency_ranges = self._frequency_range
        else:
            frequency_ranges = [self._frequency_range]


        for i in isotopologues:
            for m in models:
                for fr in frequency_ranges:
                    tss += [self._get_tag_string(z, i, m, fr)]

        return ",".join(tss)

    #
    # Jacobian & retrieval
    #

    @property
    def jacobian_class(self):
        return Jacobian

    @property
    def retrieval_class(self):
        return Retrieval

    def set_from_x(self, ws, x):

        grids = [ws.p_grid.value, ws.lat_grid.value, ws.lon_grid.value]
        grids = [g for g in grids if g.size > 0]
        x = self.transformation.invert(x)
        x = self.retrieval.interpolate_to_grids(x, grids)
        x = x.reshape(ws.vmr_field.value.shape[1:])

        unit = self.retrieval.unit
        x = unit.to_arts(ws, x)

        if self._wsv_index is None:
            raise Exception("Absorber's wsv_index is unknown. This is likely "
                            "its setup(...) routine has not been called.")

        ws.vmr_field.value[self._wsv_index, :, :, :] = x

    #
    # Retrieval
    #

    @property
    def retrieved(self):
        return not self._retrieval is None

    #
    # Abstract methods
    #

    def setup(self, ws, i):
        self._wsv_index = i

    def get_data(self, ws, provider, *args, **kwargs):

        if not self.retrieved:

            dimensions = ws.t_field.shape
            f = provider.__getattribute__("get_" + self.name)
            x = f(*args, **kwargs)
            x = extend_dimensions(x)

            if not x.shape == dimensions:
                raise Exception("Shape of {0} field is inconsistent with "
                                "the dimensions of the atmosphere."
                                .format(self.name))

            ws.vmr_field.value[self._wsv_index, :, :, :] = x

class H2O(AbsorptionSpecies):
    def __init__(self,
                 from_catalog = False,
                 cia = None,
                 frequency_range = None,
                 isotopologues = None,
                 model = "PWR98",
                 on_the_fly = True,
                 zeeman = False,
                 lineshape = "VP",
                 normalization = "VVW",
                 cutoff = -1,
                 cutoff_type = "ByBand"):
        super().__init__("H2O",
                         from_catalog = from_catalog,
                         cia = cia,
                         frequency_range = frequency_range,
                         isotopologues = isotopologues,
                         model = model,
                         on_the_fly = on_the_fly,
                         zeeman = zeeman,
                         lineshape = lineshape,
                         normalization = normalization,
                         cutoff = cutoff,
                         cutoff_type = cutoff_type)

class N2(AbsorptionSpecies):
    def __init__(self,
                 from_catalog = False,
                 cia = None,
                 frequency_range = None,
                 isotopologues = None,
                 model = "SelfContStandardType",
                 on_the_fly = True,
                 zeeman = False,
                 lineshape = "VP",
                 normalization = "VVH",
                 cutoff = -1,
                 cutoff_type = "ByBand"):
        super().__init__("N2",
                         from_catalog = from_catalog,
                         cia = cia,
                         frequency_range = frequency_range,
                         isotopologues = isotopologues,
                         model = model,
                         on_the_fly = on_the_fly,
                         zeeman = zeeman,
                         lineshape = lineshape,
                         normalization = normalization,
                         cutoff = cutoff,
                         cutoff_type = cutoff_type)

class O2(AbsorptionSpecies):
    def __init__(self,
                 from_catalog = False,
                 cia = None,
                 frequency_range = None,
                 isotopologues = None,
                 model = "PWR93",
                 on_the_fly = True,
                 zeeman = False,
                 lineshape = "VP",
                 normalization = "VVW",
                 cutoff = -1,
                 cutoff_type = "ByBand"):
        super().__init__("O2",
                         from_catalog = from_catalog,
                         cia = cia,
                         frequency_range = frequency_range,
                         isotopologues = isotopologues,
                         model = model,
                         on_the_fly = on_the_fly,
                         zeeman = zeeman,
                         lineshape = lineshape,
                         normalization = normalization,
                         cutoff = cutoff,
                         cutoff_type = cutoff_type)

class CloudWater(AbsorptionSpecies):
    def __init__(self,
                 from_catalog = False,
                 cia = None,
                 frequency_range = None,
                 isotopologues = None,
                 model = "MPM93",
                 on_the_fly = True,
                 zeeman = False,
                 lineshape = "VP",
                 normalization = "RQ",
                 cutoff = -1,
                 cutoff_type = "ByBand"):
        super().__init__("cloud_water",
                         from_catalog = from_catalog,
                         cia = cia,
                         frequency_range = frequency_range,
                         isotopologues = isotopologues,
                         model = model,
                         on_the_fly = on_the_fly,
                         zeeman = zeeman,
                         lineshape = lineshape,
                         normalization = normalization,
                         cutoff = cutoff,
                         cutoff_type = cutoff_type)

    def get_tag_string(self):

        ts = "liquidcloud"
        ts += "-"

        if self._model:
            ts += self._model
            ts += "-"

        return ts
